#!/usr/bin/env python
#
# Copyright 2007 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#



"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database


_sym_db = _symbol_database.Default()




DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n(google/appengine/base/capabilities.proto\x12\x10google.appengine\"\x86\x01\n\x14\x43\x61pabilityConfigList\x12\x32\n\x06\x63onfig\x18\x01 \x03(\x0b\x32\".google.appengine.CapabilityConfig\x12:\n\x0e\x64\x65\x66\x61ult_config\x18\x02 \x01(\x0b\x32\".google.appengine.CapabilityConfig\"\xa9\x02\n\x10\x43\x61pabilityConfig\x12\x0f\n\x07package\x18\x01 \x02(\t\x12\x12\n\ncapability\x18\x02 \x02(\t\x12\x42\n\x06status\x18\x03 \x01(\x0e\x32).google.appengine.CapabilityConfig.Status:\x07UNKNOWN\x12\x16\n\x0escheduled_time\x18\x07 \x01(\t\x12\x18\n\x10internal_message\x18\x04 \x01(\t\x12\x15\n\radmin_message\x18\x05 \x01(\t\x12\x15\n\rerror_message\x18\x06 \x01(\t\"L\n\x06Status\x12\x0b\n\x07\x44\x45\x46\x41ULT\x10\x00\x12\x0b\n\x07\x45NABLED\x10\x01\x12\r\n\tSCHEDULED\x10\x02\x12\x0c\n\x08\x44ISABLED\x10\x03\x12\x0b\n\x07UNKNOWN\x10\x04\x42:\n%com.google.appengine.api.capabilitiesB\x0e\x43\x61pabilitiesPb\xf8\x01\x01')



_CAPABILITYCONFIGLIST = DESCRIPTOR.message_types_by_name['CapabilityConfigList']
_CAPABILITYCONFIG = DESCRIPTOR.message_types_by_name['CapabilityConfig']
_CAPABILITYCONFIG_STATUS = _CAPABILITYCONFIG.enum_types_by_name['Status']
CapabilityConfigList = _reflection.GeneratedProtocolMessageType('CapabilityConfigList', (_message.Message,), {
  'DESCRIPTOR' : _CAPABILITYCONFIGLIST,
  '__module__' : 'google.appengine.base.capabilities_pb2'

  })
_sym_db.RegisterMessage(CapabilityConfigList)

CapabilityConfig = _reflection.GeneratedProtocolMessageType('CapabilityConfig', (_message.Message,), {
  'DESCRIPTOR' : _CAPABILITYCONFIG,
  '__module__' : 'google.appengine.base.capabilities_pb2'

  })
_sym_db.RegisterMessage(CapabilityConfig)

if _descriptor._USE_C_DESCRIPTORS == False:

  DESCRIPTOR._options = None
  DESCRIPTOR._serialized_options = b'\n%com.google.appengine.api.capabilitiesB\016CapabilitiesPb\370\001\001'
  _CAPABILITYCONFIGLIST._serialized_start=63
  _CAPABILITYCONFIGLIST._serialized_end=197
  _CAPABILITYCONFIG._serialized_start=200
  _CAPABILITYCONFIG._serialized_end=497
  _CAPABILITYCONFIG_STATUS._serialized_start=421
  _CAPABILITYCONFIG_STATUS._serialized_end=497




#!/usr/bin/env python
#
# Copyright 2007 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#


"""Callbacks the runtime invokes when requests end."""



from google.appengine.runtime import context



REQUEST_ID_KEY = 'HTTP_X_CLOUD_TRACE_CONTEXT'
_callback_storage = {}


def SetRequestEndCallback(callback):
  """Stores a callback by the request ID.

  The request ID currently uses the cloud trace ID.

  Args:
    callback: A zero-argument callable whose return value is unused.
  """
  req_id = GetRequestId()




  if req_id:
    _callback_storage.setdefault(req_id, []).append(callback)


def InvokeCallbacks():
  """Invokes the callbacks associated with the current request ID."""

  req_id = GetRequestId()
  if req_id in _callback_storage:
    for callback in _callback_storage[req_id]:
      callback(req_id)

    del _callback_storage[req_id]


def GetRequestId():
  """Returns a unique ID using the cloud trace ID."""
  return context.get(REQUEST_ID_KEY, None)



#!/usr/bin/env python
#
# Copyright 2007 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
















"""Models and helper functions for access to app's datastore metadata.

These entities cannot be created by users, but are created as results of
__namespace__, __kind__, __property__ and __entity_group__ metadata queries
or gets.

A simplified API is also offered:

  ndb.metadata.get_namespaces(): A list of namespace names.
  ndb.metadata.get_kinds(): A list of kind names.
  ndb.metadata.get_properties_of_kind(kind):
    A list of property names for the given kind name.
  ndb.metadata.get_representations_of_kind(kind):
    A dict mapping property names to lists of representation ids.
  ndb.metadata.get_entity_group_version(key):
    The version of the entity group containing key (HRD only).

get_kinds(), get_properties_of_kind(), get_representations_of_kind()
implicitly apply to the current namespace.

get_namespaces(), get_kinds(), get_properties_of_kind(),
get_representations_of_kind() have optional start and end arguments to limit the
query to a range of names, such that start <= name < end.
"""

from google.appengine.ext.ndb import model

__all__ = ['Namespace', 'Kind', 'Property', 'EntityGroup',
           'get_namespaces', 'get_kinds',
           'get_properties_of_kind', 'get_representations_of_kind',
           'get_entity_group_version',
          ]


class _BaseMetadata(model.Model):
  """Base class for all metadata models."""

  _use_cache = False
  _use_memcache = False

  KIND_NAME = ''

  @classmethod
  def _get_kind(cls):
    """Kind name override."""
    return cls.KIND_NAME


class Namespace(_BaseMetadata):
  """Model for __namespace__ metadata query results."""

  KIND_NAME = '__namespace__'
  EMPTY_NAMESPACE_ID = 1

  @property
  def namespace_name(self):
    """Return the namespace name specified by this entity's key."""
    return self.key_to_namespace(self.key)

  @classmethod
  def key_for_namespace(cls, namespace):
    """Return the Key for a namespace.

    Args:
      namespace: A string giving the namespace whose key is requested.

    Returns:
      The Key for the namespace.
    """
    if namespace:
      return model.Key(cls.KIND_NAME, namespace)
    else:
      return model.Key(cls.KIND_NAME, cls.EMPTY_NAMESPACE_ID)

  @classmethod
  def key_to_namespace(cls, key):
    """Return the namespace specified by a given __namespace__ key.

    Args:
      key: key whose name is requested.

    Returns:
      The namespace specified by key.
    """
    return key.string_id() or ''


class Kind(_BaseMetadata):
  """Model for __kind__ metadata query results."""

  KIND_NAME = '__kind__'

  @property
  def kind_name(self):
    """Return the kind name specified by this entity's key."""
    return self.key_to_kind(self.key)

  @classmethod
  def key_for_kind(cls, kind):
    """Return the __kind__ key for kind.

    Args:
      kind: kind whose key is requested.

    Returns:
      The key for kind.
    """
    return model.Key(cls.KIND_NAME, kind)

  @classmethod
  def key_to_kind(cls, key):
    """Return the kind specified by a given __kind__ key.

    Args:
      key: key whose name is requested.

    Returns:
      The kind specified by key.
    """
    return key.id()


class Property(_BaseMetadata):
  """Model for __property__ metadata query results."""

  KIND_NAME = '__property__'

  @property
  def property_name(self):
    """Return the property name specified by this entity's key."""
    return self.key_to_property(self.key)

  @property
  def kind_name(self):
    """Return the kind name specified by this entity's key."""
    return self.key_to_kind(self.key)

  property_representation = model.StringProperty(repeated=True)

  @classmethod
  def key_for_kind(cls, kind):
    """Return the __property__ key for kind.

    Args:
      kind: kind whose key is requested.

    Returns:
      The parent key for __property__ keys of kind.
    """
    return model.Key(Kind.KIND_NAME, kind)

  @classmethod
  def key_for_property(cls, kind, property):
    """Return the __property__ key for property of kind.

    Args:
      kind: kind whose key is requested.
      property: property whose key is requested.

    Returns:
      The key for property of kind.
    """
    return model.Key(Kind.KIND_NAME, kind, Property.KIND_NAME, property)

  @classmethod
  def key_to_kind(cls, key):
    """Return the kind specified by a given __property__ key.

    Args:
      key: key whose kind name is requested.

    Returns:
      The kind specified by key.
    """
    if key.kind() == Kind.KIND_NAME:
      return key.id()
    else:
      return key.parent().id()

  @classmethod
  def key_to_property(cls, key):
    """Return the property specified by a given __property__ key.

    Args:
      key: key whose property name is requested.

    Returns:
      property specified by key, or None if the key specified only a kind.
    """
    if key.kind() == Kind.KIND_NAME:
      return None
    else:
      return key.id()


class EntityGroup(_BaseMetadata):
  """Model for __entity_group__ metadata (available in HR datastore only).

  This metadata contains a numeric __version__ property that is guaranteed
  to increase on every change to the entity group. The version may increase
  even in the absence of user-visible changes to the entity group. The
  __entity_group__ entity may not exist if the entity group was never
  written to.
  """

  KIND_NAME = '__entity_group__'
  ID = 1

  version = model.IntegerProperty(name='__version__')

  @classmethod
  def key_for_entity_group(cls, key):
    """Return the key for the entity group containing key.

    Args:
      key: a key for an entity group whose __entity_group__ key you want.

    Returns:
      The __entity_group__ key for the entity group containing key.
    """
    return model.Key(cls.KIND_NAME, cls.ID, parent=key.root())


def get_namespaces(start=None, end=None):
  """Return all namespaces in the specified range.

  Args:
    start: only return namespaces >= start if start is not None.
    end: only return namespaces < end if end is not None.

  Returns:
    A list of namespace names between the (optional) start and end values.
  """
  q = Namespace.query()
  if start is not None:
    q = q.filter(Namespace.key >= Namespace.key_for_namespace(start))
  if end is not None:
    q = q.filter(Namespace.key < Namespace.key_for_namespace(end))
  return [x.namespace_name for x in q]


def get_kinds(start=None, end=None):
  """Return all kinds in the specified range, for the current namespace.

  Args:
    start: only return kinds >= start if start is not None.
    end: only return kinds < end if end is not None.

  Returns:
    A list of kind names between the (optional) start and end values.
  """
  q = Kind.query()
  if start is not None and start != '':
    q = q.filter(Kind.key >= Kind.key_for_kind(start))
  if end is not None:
    if end == '':
      return []
    q = q.filter(Kind.key < Kind.key_for_kind(end))

  return [x.kind_name for x in q]


def get_properties_of_kind(kind, start=None, end=None):
  """Return all properties of kind in the specified range.

  NOTE: This function does not return unindexed properties.

  Args:
    kind: name of kind whose properties you want.
    start: only return properties >= start if start is not None.
    end: only return properties < end if end is not None.

  Returns:
    A list of property names of kind between the (optional) start and end
    values.
  """
  q = Property.query(ancestor=Property.key_for_kind(kind))
  if start is not None and start != '':
    q = q.filter(Property.key >= Property.key_for_property(kind, start))
  if end is not None:
    if end == '':
      return []
    q = q.filter(Property.key < Property.key_for_property(kind, end))

  return [Property.key_to_property(k) for k in q.iter(keys_only=True)]


def get_representations_of_kind(kind, start=None, end=None):
  """Return all representations of properties of kind in the specified range.

  NOTE: This function does not return unindexed properties.

  Args:
    kind: name of kind whose properties you want.
    start: only return properties >= start if start is not None.
    end: only return properties < end if end is not None.

  Returns:
    A dictionary mapping property names to its list of representations.
  """
  q = Property.query(ancestor=Property.key_for_kind(kind))
  if start is not None and start != '':
    q = q.filter(Property.key >= Property.key_for_property(kind, start))
  if end is not None:
    if end == '':
      return {}
    q = q.filter(Property.key < Property.key_for_property(kind, end))

  result = {}
  for property in q:
    result[property.property_name] = property.property_representation

  return result


def get_entity_group_version(key):
  """Return the version of the entity group containing key.

  Args:
    key: a key for an entity group whose __entity_group__ key you want.

  Returns:
    The version of the entity group containing key. This version is
    guaranteed to increase on every change to the entity group. The version
    may increase even in the absence of user-visible changes to the entity
    group. May return None if the entity group was never written to.
  """

  eg = EntityGroup.key_for_entity_group(key).get()
  if eg:
    return eg.version
  else:
    return None



#!/usr/bin/env python
#
# Copyright 2007 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
r"""An Encoder class for protocol buffers (PB) that preserves sorting characteristics.

This is used by `datastore_sqlite_stub` and `datastore_types` to match the
ordering semantics of the production datastore. Broadly, there are four
changes from regular PB encoding:

 - Strings are escaped and null terminated instead of length-prefixed. The
   escaping replaces `\x00` with `\x01\x01` and `\x01` with `\x01\x02`, thus
   preserving the ordering of the original string.
 - Variable length integers are encoded using a variable length encoding that
   preserves order. The first byte stores the absolute value if it's between
   -119 to 119, otherwise it stores the number of bytes that follow.
 - Numbers are stored big endian instead of little endian.
 - Negative doubles are entirely negated, while positive doubles have their sign
   bit flipped.

Warning:
  Due to the way nested Protocol Buffers are encoded, this encoder will NOT
  preserve sorting characteristics for embedded protocol buffers!
"""




import array
import struct
import six

from google.protobuf import descriptor
from google.protobuf import message
from google.protobuf import symbol_database

_MAX_UNSIGNED_BYTE = 255


_MAX_LONG_BYTES = 8




_MAX_INLINE = (_MAX_UNSIGNED_BYTE - (2 * _MAX_LONG_BYTES)) // 2
_MIN_INLINE = -_MAX_INLINE
_OFFSET = 1 + 8
_POS_OFFSET = _OFFSET + _MAX_INLINE * 2


def ToBytes(arr):
  if six.PY2:
    return arr.tostring()
  else:
    return arr.tobytes()


def FromBytes(arr, b):
  if six.PY2:
    arr.fromstring(b)
  else:
    arr.frombytes(b)


def GetFieldNumber(field_tuple):
  """Get the field number from the tuple returned by `Message.ListFields`."""
  return field_tuple[0].number


_GROUP_START_WIRE_TYPE = 3
_GROUP_END_WIRE_TYPE = 4

_WIRE_TYPES = {
    descriptor.FieldDescriptor.TYPE_DOUBLE: 1,
    descriptor.FieldDescriptor.TYPE_FLOAT: 5,
    descriptor.FieldDescriptor.TYPE_INT64: 0,
    descriptor.FieldDescriptor.TYPE_UINT64: 0,
    descriptor.FieldDescriptor.TYPE_INT32: 0,
    descriptor.FieldDescriptor.TYPE_FIXED64: 1,
    descriptor.FieldDescriptor.TYPE_FIXED32: 5,
    descriptor.FieldDescriptor.TYPE_BOOL: 0,
    descriptor.FieldDescriptor.TYPE_STRING: 2,
    descriptor.FieldDescriptor.TYPE_MESSAGE: 2,
    descriptor.FieldDescriptor.TYPE_GROUP: 3,
    descriptor.FieldDescriptor.TYPE_BYTES: 3,
}


class Encoder(object):
  """Encodes protocol buffers in a form that sorts nicely."""

  @classmethod
  def EncodeMessage(cls, msg):
    encoder = cls()
    encoder.PutMessage(msg)
    return six.ensure_binary(ToBytes(encoder.Buffer()))

  def __init__(self):
    self.buf = array.array('B')
    return

  def Buffer(self):
    return self.buf

  def Put16(self, value):
    """Encode a fixed size 16 bit value in the buffer."""
    if value < 0 or value >= (1 << 16):
      raise message.EncodeError('u16 too big')
    self.buf.append((value >> 8) & 0xff)
    self.buf.append((value >> 0) & 0xff)
    return

  def Put32(self, value):
    """Encode a fixed size 32 bit value in the buffer."""
    if value < 0 or value >= (1 << 32):
      raise message.EncodeError('u32 too big')
    self.buf.append((value >> 24) & 0xff)
    self.buf.append((value >> 16) & 0xff)
    self.buf.append((value >> 8) & 0xff)
    self.buf.append((value >> 0) & 0xff)
    return

  def Put64(self, value):
    """Encode a fixed size 64 bit value in the buffer."""
    if value < 0 or value >= (1 << 64):
      raise message.EncodeError('u64 too big')
    self.buf.append((value >> 56) & 0xff)
    self.buf.append((value >> 48) & 0xff)
    self.buf.append((value >> 40) & 0xff)
    self.buf.append((value >> 32) & 0xff)
    self.buf.append((value >> 24) & 0xff)
    self.buf.append((value >> 16) & 0xff)
    self.buf.append((value >> 8) & 0xff)
    self.buf.append((value >> 0) & 0xff)
    return

  def _PutVarInt(self, value):
    """Encode a `varint` value in the buffer."""
    if value is None:
      self.buf.append(0)
      return

    if value >= _MIN_INLINE and value <= _MAX_INLINE:
      value = _OFFSET + (value - _MIN_INLINE)
      self.buf.append(value & 0xff)
      return

    negative = False

    if value < 0:
      value = _MIN_INLINE - value
      negative = True
    else:
      value = value - _MAX_INLINE

    length = 0
    w = value
    while w > 0:
      w >>= 8
      length += 1

    if negative:
      head = _OFFSET - length
    else:
      head = _POS_OFFSET + length
    self.buf.append(head & 0xff)

    for i in range(length - 1, -1, -1):
      b = value >> (i * 8)
      if negative:
        b = _MAX_UNSIGNED_BYTE - (b & 0xff)
      self.buf.append(b & 0xff)

  def PutVarInt32(self, value):
    """Encode a 32 bit `varint` value in the buffer."""
    if value >= 0x80000000 or value < -0x80000000:
      raise message.EncodeError('int32 too big')
    self._PutVarInt(value)

  def PutVarInt64(self, value):
    """Encode a 64 bit `varint` value in the buffer."""
    if value >= 0x8000000000000000 or value < -0x8000000000000000:
      raise message.EncodeError('int64 too big')
    self._PutVarInt(value)

  def PutVarUint64(self, value):
    """Encode a 64 bit unsigned `varint` value in the buffer."""
    if value < 0 or value >= 0x10000000000000000:
      raise message.EncodeError('uint64 too big')
    self._PutVarInt(value)

  def PutFloat(self, value):
    """Encode a floating point value in the buffer."""
    encoded = array.array('B')
    FromBytes(encoded, struct.pack('>f', value))
    if IsFloatNegative(value, encoded):


      encoded[0] ^= 0xFF
      encoded[1] ^= 0xFF
      encoded[2] ^= 0xFF
      encoded[3] ^= 0xFF
    else:

      encoded[0] ^= 0x80
    self.buf.extend(encoded)

  def PutDouble(self, value):
    """Encode a double value in the buffer."""
    encoded = array.array('B')
    FromBytes(encoded, struct.pack('>d', value))
    if IsFloatNegative(value, encoded):


      encoded[0] ^= 0xFF
      encoded[1] ^= 0xFF
      encoded[2] ^= 0xFF
      encoded[3] ^= 0xFF
      encoded[4] ^= 0xFF
      encoded[5] ^= 0xFF
      encoded[6] ^= 0xFF
      encoded[7] ^= 0xFF
    else:

      encoded[0] ^= 0x80
    self.buf.extend(encoded)

  def PutPrefixedString(self, value):


    null_terminated = value.replace('\x01', '\x01\x02').replace(
        '\x00', '\x01\x01') + '\x00'
    FromBytes(self.buf, six.ensure_binary(null_terminated))

  def PutBytes(self, value):
    """Encode a byte array in the buffer."""
    null_terminated = value.replace(
        six.ensure_binary('\x01'), six.ensure_binary('\x01\x02')).replace(
            six.ensure_binary('\x00'),
            six.ensure_binary('\x01\x01')) + six.ensure_binary('\x00')
    FromBytes(self.buf, six.ensure_binary(null_terminated))

  def PutBoolean(self, v):
    """Encode a boolean value in the buffer."""
    if v:
      self.buf.append(1)
    else:
      self.buf.append(0)
    return

  def PutMessage(self, msg):
    """Encode a protocol buffer message on the buffer."""
    for field_descriptor, value in sorted(msg.ListFields(), key=GetFieldNumber):
      if field_descriptor.label == descriptor.FieldDescriptor.LABEL_REPEATED:
        self._OutputRepeatedValue(field_descriptor, value)
      else:
        self._PutScalarValue(field_descriptor, value)

  def _OutputRepeatedValue(self, field_descriptor, value):
    if field_descriptor.GetOptions().packed:
      raise NotImplementedError('Packed values not supported')
    for v in value:
      self._PutScalarValue(field_descriptor, v)

  def _PutFieldTag(self, field_descriptor):
    self.PutVarInt32((field_descriptor.number << 3)
                     | _WIRE_TYPES[field_descriptor.type])

  def _PutEndGroupTag(self, field_descriptor):
    self.PutVarInt32((field_descriptor.number << 3) | 4)

  def _PutScalarValue(self, field_descriptor, value):
    """Encode a scalar value onto the buffer."""
    self._PutFieldTag(field_descriptor)
    if field_descriptor.type == descriptor.FieldDescriptor.TYPE_DOUBLE:
      self.PutDouble(value)
    elif field_descriptor.type == descriptor.FieldDescriptor.TYPE_FLOAT:
      self.PutFloat(value)
    elif field_descriptor.type == descriptor.FieldDescriptor.TYPE_INT64:
      self.PutVarInt64(value)
    elif field_descriptor.type == descriptor.FieldDescriptor.TYPE_UINT64:
      self.PutVarUint64(value)
    elif field_descriptor.type == descriptor.FieldDescriptor.TYPE_INT32:
      self.PutVarInt32(value)
    elif field_descriptor.type == descriptor.FieldDescriptor.TYPE_FIXED64:
      self.Put64(value)
    elif field_descriptor.type == descriptor.FieldDescriptor.TYPE_FIXED32:
      self.Put32(value)
    elif field_descriptor.type == descriptor.FieldDescriptor.TYPE_BOOL:
      self.PutBoolean(value)
    elif field_descriptor.type == descriptor.FieldDescriptor.TYPE_STRING:
      self.PutPrefixedString(value)
    elif field_descriptor.type == descriptor.FieldDescriptor.TYPE_GROUP:
      self.PutMessage(value)
      self._PutEndGroupTag(field_descriptor)
    elif field_descriptor.type == descriptor.FieldDescriptor.TYPE_MESSAGE:
      encoded_bytes = self.EncodeMessage(value)
      self.PutBytes(encoded_bytes)
    elif field_descriptor.type == descriptor.FieldDescriptor.TYPE_BYTES:
      self.PutBytes(value)
    else:
      error_message = ('Unsupported field type. Name: %s Type: %d' %
                       (field_descriptor.full_name, field_descriptor.type))
      raise NotImplementedError(error_message)


class Decoder(object):
  """Decoder that decodes a bytes buffer into a protocol buffer message."""

  def __init__(self, buf, idx=0, limit=None):
    self.buf = buf
    self.idx = idx
    self.limit = limit or len(buf)

  @classmethod
  def DecodeMessage(cls, prototype, buf, idx=0, limit=None):
    return cls(buf, idx, limit).GetMessage(prototype)

  def _Avail(self):
    return self.limit - self.idx

  def _Buffer(self):
    return self.buf

  def _Pos(self):
    return self.idx

  def _Skip(self, n):
    if self.idx + n > self.limit:
      raise message.DecodeError('truncated')
    self.idx += n
    return

  def Get8(self):
    """Get the next 8 bits from the buffer."""
    if self.idx >= self.limit:
      raise message.DecodeError('truncated')
    c = self.buf[self.idx]
    self.idx += 1
    return c

  def Get16(self):
    """Get the next 16 bits from the buffer."""
    if self.idx + 2 > self.limit:
      raise message.DecodeError('truncated')
    c = self.buf[self.idx]
    d = self.buf[self.idx + 1]
    self.idx += 2
    return (c << 8) | d

  def Get32(self):
    """Get the next 32 bits from the buffer."""
    if self.idx + 4 > self.limit:
      raise message.DecodeError('truncated')
    c = int(self.buf[self.idx])
    d = self.buf[self.idx + 1]
    e = self.buf[self.idx + 2]
    f = self.buf[self.idx + 3]
    self.idx += 4
    return (c << 24) | (d << 16) | (e << 8) | f

  def Get64(self):
    """Get the next 64 bits from the buffer."""
    if self.idx + 8 > self.limit:
      raise message.DecodeError('truncated')
    c = int(self.buf[self.idx])
    d = int(self.buf[self.idx + 1])
    e = int(self.buf[self.idx + 2])
    f = int(self.buf[self.idx + 3])
    g = int(self.buf[self.idx + 4])
    h = self.buf[self.idx + 5]
    i = self.buf[self.idx + 6]
    j = self.buf[self.idx + 7]
    self.idx += 8
    return ((c << 56) | (d << 48) | (e << 40) | (f << 32) | (g << 24)
            | (h << 16) | (i << 8) | j)

  def GetVarInt64(self):
    """Decode the next bits as a `varint`."""
    b = self.Get8()
    if b >= _OFFSET and b <= _POS_OFFSET:
      return b - _OFFSET + _MIN_INLINE
    if b == 0:
      return None

    if b < _OFFSET:
      negative = True
      num_bytes = _OFFSET - b
    else:
      negative = False
      num_bytes = b - _POS_OFFSET

    ret = 0
    for _ in range(num_bytes):
      b = self.Get8()
      if negative:
        b = _MAX_UNSIGNED_BYTE - b
      ret = ret << 8 | b

    if negative:
      return _MIN_INLINE - ret
    else:
      return ret + _MAX_INLINE

  def GetVarInt32(self):
    """Decode the next bits as a `varint` and check it's a 32 bit value."""
    result = self.GetVarInt64()
    if result >= 0x80000000 or result < -0x80000000:
      raise message.DecodeError('corrupted')
    return result

  def GetVarUint64(self):
    """Decode the next bits as a `varint` and check it's non-negative value."""
    result = self.GetVarInt64()
    if result < 0:
      raise message.DecodeError('corrupted')
    return result

  def GetFloat(self):
    """Decode the next bits as a floating point value."""
    if self.idx + 4 > self.limit:
      raise message.DecodeError('truncated')
    a = self.buf[self.idx:self.idx + 4]
    self.idx += 4
    if a[0] & 0x80:

      a[0] ^= 0x80
    else:

      a = [x ^ 0xFF for x in a]
    return struct.unpack('>f', ToBytes(array.array('B', a)))[0]

  def GetDouble(self):
    """Decode the next bits as a double value."""
    if self.idx + 8 > self.limit:
      raise message.DecodeError('truncated')
    a = self.buf[self.idx:self.idx + 8]
    self.idx += 8
    if a[0] & 0x80:

      a[0] ^= 0x80
    else:

      a = [x ^ 0xFF for x in a]
    return struct.unpack('>d', ToBytes(array.array('B', a)))[0]

  def GetPrefixedString(self):
    end_idx = self.idx
    while self.buf[end_idx] != 0:
      end_idx += 1

    data = ToBytes(array.array('B', self.buf[self.idx:end_idx]))
    self.idx = end_idx + 1
    return data.replace(
        six.ensure_binary('\x01\x01'), six.ensure_binary('\x00')).replace(
            six.ensure_binary('\x01\x02'), six.ensure_binary('\x01'))

  def GetBoolean(self):
    """Decode the next bits as a boolean value."""
    b = self.Get8()
    if b != 0 and b != 1:
      raise message.DecodeError('corrupted')
    return b

  def _GetNumAndWireType(self):
    tt = self.GetVarInt32()
    return tt >> 3, tt & 7

  def _GetValue(self, field_descriptor, wire_type):
    """Get the next encoded value based on the field information."""
    if field_descriptor.type == descriptor.FieldDescriptor.TYPE_DOUBLE:
      return self.GetDouble()
    elif field_descriptor.type == descriptor.FieldDescriptor.TYPE_FLOAT:
      return self.GetFloat()
    elif field_descriptor.type == descriptor.FieldDescriptor.TYPE_INT64:
      return self.GetVarInt64()
    elif field_descriptor.type == descriptor.FieldDescriptor.TYPE_UINT64:
      return self.GetVarUint64()
    elif field_descriptor.type == descriptor.FieldDescriptor.TYPE_INT32:
      return self.GetVarInt32()
    elif field_descriptor.type == descriptor.FieldDescriptor.TYPE_FIXED64:
      return self.Get64()
    elif field_descriptor.type == descriptor.FieldDescriptor.TYPE_FIXED32:
      return self.Get32()
    elif field_descriptor.type == descriptor.FieldDescriptor.TYPE_BOOL:
      return self.GetBoolean()
    elif field_descriptor.type == descriptor.FieldDescriptor.TYPE_STRING:
      return six.ensure_text(self.GetPrefixedString())
    elif field_descriptor.type in (descriptor.FieldDescriptor.TYPE_GROUP,
                                   descriptor.FieldDescriptor.TYPE_MESSAGE):
      prototype = _GetPrototype(field_descriptor)
      if wire_type == _GROUP_START_WIRE_TYPE:
        return self.GetMessage(prototype)
      else:
        arr = array.array('B')
        FromBytes(arr, self.GetPrefixedString())
        sub_decoder = Decoder(arr, 0, len(arr))
        return sub_decoder.GetMessage(prototype)
    elif field_descriptor.type == descriptor.FieldDescriptor.TYPE_BYTES:
      return self.GetPrefixedString()
    else:
      error_message = ('Unsupported field type. Name: %s Type: %d' %
                       (field_descriptor.full_name, field_descriptor.type))
      raise NotImplementedError(error_message)

  def GetMessage(self, prototype):
    """Decode a message.

    Args:
      prototype: The prototype of the protocol buffer message.

    Returns:
      The decoded message.
    """

    msg = prototype()
    message_descriptor = prototype.DESCRIPTOR

    while self._Avail() > 0:
      number, wire_type = self._GetNumAndWireType()
      if wire_type == _GROUP_END_WIRE_TYPE:
        break
      field_descriptor = message_descriptor.fields_by_number[number]
      if field_descriptor.GetOptions().packed:
        raise NotImplementedError('Packed values not supported.')
      value = self._GetValue(field_descriptor, wire_type)

      if field_descriptor.label == descriptor.FieldDescriptor.LABEL_REPEATED:
        getattr(msg, field_descriptor.name).append(value)
      else:
        setattr(msg, field_descriptor.name, value)

    return msg


def _GetPrototype(field_descriptor):
  return symbol_database.Default().GetPrototype(field_descriptor.message_type)


def IsFloatNegative(value, encoded):
  if value == 0:
    return encoded[0] == 128
  return value < 0


def EncodeDouble(value):
  """Encode a double into a sortable byte buffer."""

  encoded = array.array('B')
  FromBytes(encoded, struct.pack('>d', value))
  if IsFloatNegative(value, encoded):


    encoded[0] ^= 0xFF
    encoded[1] ^= 0xFF
    encoded[2] ^= 0xFF
    encoded[3] ^= 0xFF
    encoded[4] ^= 0xFF
    encoded[5] ^= 0xFF
    encoded[6] ^= 0xFF
    encoded[7] ^= 0xFF
  else:

    encoded[0] ^= 0x80
  return encoded



#!/usr/bin/env python
#
# Copyright 2007 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

"""A module that handles deferred execution of callables via the task queue.

Tasks consist of a callable and arguments to pass to it. The callable and its
arguments are serialized and put on the task queue, which deserializes and
executes them. The following callables can be used as tasks:

1) Functions defined in the top level of a module
2) Classes defined in the top level of a module
3) Instances of classes in (2) that implement __call__
4) Instance methods of objects of classes in (2)
5) Class methods of classes in (2)
6) Built-in functions
7) Built-in methods

The following callables can NOT be used as tasks:
1) Nested functions or closures
2) Nested classes or objects of them
3) Lambda functions
4) Static methods

The arguments to the callable, and the object (in the case of method or object
calls) must all be pickleable.

If you want your tasks to execute reliably, don't use mutable global variables;
they are not serialized with the task and may not be the same when your task
executes as they were when it was enqueued (in fact, they will almost certainly
be different).

If your app relies on manipulating the import path, make sure that the function
you are deferring is defined in a module that can be found without import path
manipulation. Alternately, you can include deferred.TaskHandler in your own
webapp application instead of using the easy-install method detailed below.

When you create a deferred task using deferred.defer, the task is serialized,
and an attempt is made to add it directly to the task queue. If the task is too
big (larger than about 10 kilobytes when serialized), a datastore entry will be
created for the task, and a new task will be enqueued, which will fetch the
original task from the datastore and execute it. This is much less efficient
than the direct execution model, so it's a good idea to minimize the size of
your tasks when possible.

By default, the deferred module uses the URL '/_ah/queue/deferred',
and the default queue.

To enable the Deferred API, set 'use_deferred=True' in the call
to 'wrap_wsgi_app()'.

Example for a Flask app:

app = Flask(__name__)
app.wsgi_app = wrap_wsgi_app(app.wsgi_app, use_deferred=True)

Deferring a task in Flask:

```
from flask import Flask, request
from google.appengine.api import wrap_wsgi_app
from google.appengine.ext import ndb
from google.appengine.ext import deferred

class MyModel(ndb.Model):
  total = ndb.IntegerProperty(indexed=True)

my_key = "defaultKey"

def do_something_later(key, amount):
   entity = MyModel.get_or_insert(key, total=0)
   entity.total += amount
   entity.put()

@app.route("/home")
def hello_world():
  # Use default URL and queue name, no task name, execute ASAP.
  deferred.defer(do_something_later, my_key, 20)

  # Execute after 60s
  deferred.defer(do_something_later, my_key, 20, _countdown=60)

  # Using a non-default queue (a TaskQueue 'foo' should already exist)
  deferred.defer(do_something_later, my_key, 20, _queue="foo", _countdown=60)

app = Flask(__name__)
app.wsgi_app = wrap_wsgi_app(app.wsgi_app, use_deferred=True)
```
"""








import http
import logging
import pickle
import types
from google.appengine.api import taskqueue
from google.appengine.ext import db


_DEFAULT_LOG_LEVEL = logging.INFO
_TASKQUEUE_HEADERS = {"Content-Type": "application/octet-stream"}
_TASKQUEUE_RESPONSE_HEADERS = ("Content-Type", "text/plain")
_DEFAULT_URL = "/_ah/queue/deferred"
_DEFAULT_QUEUE = "default"


class Error(Exception):
  """Base class for exceptions in this module."""


class PermanentTaskFailure(Error):
  """Indicates that a task failed, and will never succeed."""


class SingularTaskFailure(Error):
  """Indicates that a task failed once."""


def set_log_level(log_level):
  """Sets the log level deferred will log to in normal circumstances.

  Args:
    log_level: one of logging log levels, e.g. logging.DEBUG, logging.INFO, etc.
  """
  global _DEFAULT_LOG_LEVEL
  _DEFAULT_LOG_LEVEL = log_level


def run(data):
  """Unpickles and executes a task.

  Args:
    data: A pickled tuple of (function, args, kwargs) to execute.

  Returns:
    The return value of the function invocation.

  Raises:
    PermanentTaskFailure if an error occurred during unpickling the task.
  """
  try:
    func, args, kwds = pickle.loads(data)
  except Exception as e:
    raise PermanentTaskFailure(e)
  else:
    return func(*args, **kwds)


class _DeferredTaskEntity(db.Model):
  """Datastore representation of a deferred task.

  This is used in cases when the deferred task is too big to be included as
  payload with the task queue entry.
  """
  data = db.BlobProperty(required=True)


def run_from_datastore(key):
  """Retrieves a task from the datastore and executes it.

  Args:
    key: The datastore key of a _DeferredTaskEntity storing the task.

  Returns:
    The return value of the function invocation.

  Raises:
    PermanentTaskFailure: Raised if the task entity is missing.
  """
  entity = _DeferredTaskEntity.get(key)
  if not entity:

    raise PermanentTaskFailure()
  try:
    ret = run(entity.data)
    entity.delete()
  except PermanentTaskFailure:
    entity.delete()
    raise


def invoke_member(obj, membername, *args, **kwargs):
  """Retrieves a member of an object, then calls it with the provided arguments.

  Args:
    obj: The object to operate on.
    membername: The name of the member to retrieve from ojb.
    *args: Positional arguments to pass to the method.
    **kwargs: Keyword arguments to pass to the method.

  Returns:
    The return value of the method invocation.
  """
  return getattr(obj, membername)(*args, **kwargs)


def _curry_callable(obj, *args, **kwargs):
  """Takes a callable and arguments and returns a task queue tuple.

  The returned tuple consists of (callable, args, kwargs), and can be pickled
  and unpickled safely.

  Args:
    obj: The callable to curry. See the module docstring for restrictions.
    *args: Positional arguments to call the callable with.
    **kwargs: Keyword arguments to call the callable with.

  Returns:
    A tuple consisting of  (callable, args, kwargs) that can be evaluated by
    run() with equivalent effect of executing the function directly.
  Raises:
    ValueError: If the passed in object is not of a valid callable type.
  """
  if isinstance(obj, types.MethodType):
    return (invoke_member, (obj.__self__, obj.__func__.__name__) + args, kwargs)
  elif isinstance(obj, types.BuiltinMethodType):
    if not obj.__self__:

      return (obj, args, kwargs)
    else:


      if isinstance(obj.__self__, types.ModuleType):
        return (obj, args, kwargs)
      else:
        return (invoke_member, (obj.__self__, obj.__name__) + args, kwargs)
  elif isinstance(obj, object) and hasattr(obj, "__call__"):
    return (obj, args, kwargs)
  elif isinstance(obj, (types.FunctionType, types.BuiltinFunctionType, type,
                        types.UnboundMethodType)):
    return (obj, args, kwargs)
  else:
    raise ValueError("obj must be callable")


def serialize(obj, *args, **kwargs):
  """Serializes a callable into a format recognized by the deferred executor.

  Args:
    obj: The callable to serialize. See module docstring for restrictions.
    *args: Positional arguments to call the callable with.
    **kwargs: Keyword arguments to call the callable with.

  Returns:
    A serialized representation of the callable.
  """
  curried = _curry_callable(obj, *args, **kwargs)
  return pickle.dumps(curried, protocol=pickle.HIGHEST_PROTOCOL)


def defer(obj, *args, **kwargs):
  """Defers a callable for execution later.

  The default deferred URL of /_ah/queue/deferred will be used unless an
  alternate URL is explicitly specified. If you want to use the default URL for
  a queue, specify _url=None. If you specify a different URL, you will need to
  install the handler on that URL (see the module docstring for details).

  Args:
    obj: The callable to execute. See module docstring for restrictions.
      _countdown, _eta, _headers, _name, _target, _transactional, _url,
        _retry_options, _queue: Passed through to the task queue - see the task
          queue documentation for details.
    *args: Positional arguments to call the callable with.
    **kwargs: Any other keyword arguments are passed through to the callable.

  Returns:
    A taskqueue.Task object which represents an enqueued callable.
  """
  taskargs = dict(
      (x, kwargs.pop(("_%s" % x), None))
      for x in ("countdown", "eta", "name", "target", "retry_options"))
  taskargs["url"] = kwargs.pop("_url", _DEFAULT_URL)
  transactional = kwargs.pop("_transactional", False)
  taskargs["headers"] = dict(_TASKQUEUE_HEADERS)
  taskargs["headers"].update(kwargs.pop("_headers", {}))
  queue = kwargs.pop("_queue", _DEFAULT_QUEUE)
  pickled = serialize(obj, *args, **kwargs)
  try:
    task = taskqueue.Task(payload=pickled, **taskargs)
    return task.add(queue, transactional=transactional)
  except taskqueue.TaskTooLargeError:

    key = _DeferredTaskEntity(data=pickled).put()
    pickled = serialize(run_from_datastore, str(key))
    task = taskqueue.Task(payload=pickled, **taskargs)
    return task.add(queue)


class Handler():
  """A handler class for processesing deferred invocations."""

  def run_from_request(self, environ):
    """Executes deferred tasks after verifying the caller.

    This function assumes that the WSGI environ dict originated from a POST
    request by the GAE TaskQueue service. It checks the caller IP and request
    headers to verify the caller.
    Args:
      environ: a WSGI dict describing the HTTP request (See PEP 333).
    Returns:
      response: a string containing body of the response
      status: HTTP status code of enum type http.HTTPStatus
      headers: a dict containing response headers
    Raises:
      PermanentTaskFailure if an error occurred during unpickling the task.
    """

    if "HTTP_X_APPENGINE_TASKNAME" not in environ:
      error_message = ("Detected an attempted XSRF attack. "
                       "The header 'X-AppEngine-Taskname' was not set.")
      logging.error(error_message)
      return error_message, http.HTTPStatus.FORBIDDEN, [
          _TASKQUEUE_RESPONSE_HEADERS
      ]



    in_prod = (
        not environ.get("SERVER_SOFTWARE").startswith("Devel"))
    if in_prod and environ.get("REMOTE_ADDR") != "0.1.0.2":
      error_message = ("Detected an attempted XSRF attack. "
                       "This request did not originate from Task Queue.")
      logging.error(error_message)
      return error_message, http.HTTPStatus.FORBIDDEN, [
          _TASKQUEUE_RESPONSE_HEADERS
      ]


    headers = [
        "%s:%s" % (k[5:], v)
        for k, v in environ.items()
        if k.upper().startswith("HTTP_X_APPENGINE_")
    ]
    logging.log(_DEFAULT_LOG_LEVEL, ", ".join(headers))


    try:
      request_body_size = int(environ.get("CONTENT_LENGTH", 0))
    except ValueError:
      request_body_size = 0

    request_body = environ["wsgi.input"].read(request_body_size)
    run(request_body)
    return "Success", http.HTTPStatus.OK, [_TASKQUEUE_RESPONSE_HEADERS]

  def post(self, environ):
    """Default behavior for POST requests to the deferred endpoint.

    If the Deferred API has been enabled, this function is automatically called
    after 'deferred.defer()' is used to defer a task. Behind the scenes,
    the TaskQueue service calls the default Deferred endpoint
    '/_ah/queue/deferred', which is routed to this function.

    If deferred.defer() is passed a custom '_url' parameter, POST requests to
    the custom endpoint needs to be handled by the app. To replicate the
    default behavior of executing a deferred task, this function can be called
    by the custom endpoint handler, and passed the WSGI 'environ' dictionary
    of the POST request.

    Args:
      environ: a WSGI dict describing the HTTP request (See PEP 333)
    Returns:
      response: a string containing body of the response
      status: HTTP status code of enum type http.HTTPStatus
      headers: a dict containing response headers
    """
    try:
      response, status, headers = self.run_from_request(environ)
    except SingularTaskFailure:


      response, status, headers = ("SingularTaskFailure",
                                   http.HTTPStatus.REQUEST_TIMEOUT,
                                   [_TASKQUEUE_RESPONSE_HEADERS])
      logging.debug("Failure executing task, task retry forced")
    except PermanentTaskFailure:

      response, status, headers = ("PermanentTaskFailure", http.HTTPStatus.OK,
                                   [_TASKQUEUE_RESPONSE_HEADERS])
      logging.exception("Permanent failure attempting to execute task")
    return response, status, headers

  def __call__(self, environ, start_response):
    """WSGI app callable to handle POST requests to the deferred endpoint.

    This function allows the 'Handler' class to behave like a WSGI app for
    handling Deferred task execution.

    If the parent WSGI app is required to handle multiple kinds of requests,
    a dispatcher (compliant with the app's Web Framework) can be used to route
    an endpoint/URL to be handled by this callable.

    Args:
      environ: a WSGI dict describing the HTTP request (See PEP 333)
      start_response: callable (See PEP 3333)
    Returns:
      list of bytes response
    """
    if environ["REQUEST_METHOD"] != "POST":
      return ("", http.HTTPStatus.METHOD_NOT_ALLOWED, [("Allow", "POST")])

    response, status, headers = self.post(environ)
    start_response(f"{status.value} {status.phrase}", headers)
    return [response.encode("utf-8")]


application = Handler()



#!/usr/bin/env python
#
# Copyright 2007 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#


"""Task Queue API module."""

from google.appengine.api.taskqueue.taskqueue import *


__all__ = [

    'BadTaskStateError',
    'BadTransactionState',
    'BadTransactionStateError',
    'DatastoreError',
    'DuplicateTaskNameError',
    'Error',
    'InternalError',
    'InvalidQueueError',
    'InvalidQueueNameError',
    'InvalidTaskError',
    'InvalidTaskNameError',
    'InvalidUrlError',
    'PermissionDeniedError',
    'TaskAlreadyExistsError',
    'TaskTooLargeError',
    'TombstonedTaskError',
    'TooManyTasksError',
    'TransientError',
    'UnknownQueueError',
    'InvalidLeaseTimeError',
    'InvalidMaxTasksError',
    'InvalidDeadlineError',
    'InvalidQueueModeError',
    'TransactionalRequestTooLargeError',
    'TaskLeaseExpiredError',
    'QueuePausedError',
    'InvalidEtaError',
    'InvalidTagError',
    'InvalidDispatchDeadlineError',

    'MAX_QUEUE_NAME_LENGTH',
    'MAX_TASK_NAME_LENGTH',
    'MAX_TASK_SIZE_BYTES',
    'MAX_PULL_TASK_SIZE_BYTES',
    'MAX_PUSH_TASK_SIZE_BYTES',
    'MAX_LEASE_SECONDS',
    'MAX_TASKS_PER_ADD',
    'MAX_TASKS_PER_LEASE',
    'MAX_URL_LENGTH',

    'DEFAULT_APP_VERSION',

    'Queue',
    'QueueStatistics',
    'Task',
    'TaskRetryOptions',
    'add',
    'create_rpc'
]




#!/usr/bin/env python
#
# Copyright 2007 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

"""Directly processes text of `datastore-indexes.xml`.

`IndexesXmlParser` is called with an XML string to produce an `IndexXml` object
containing the data from the XML.

- `IndexesXmlParser` converts XML to Index object.
- `Index` describes a single index specified in `datastore-indexes.xml`.
"""







from xml.etree import ElementTree

from google.appengine.api.validation import ValidationError
from google.appengine.datastore.datastore_index import Index
from google.appengine.datastore.datastore_index import IndexDefinitions
from google.appengine.datastore.datastore_index import Property

MISSING_KIND = '<datastore-index> node has missing attribute "kind".'
BAD_DIRECTION = ('<property> tag attribute "direction" must have value "asc"'
                 ' or "desc", given "%s"')
BAD_MODE = ('<property> tag attribute "mode" must have value "geospatial",'
            ' given "%s"')
NAME_MISSING = ('<datastore-index> node with kind "%s" needs to have a name'
                ' attribute specified for its <property> node')
MODE_AND_DIRECTION_SPECIFIED = ('<datastore-index> node has both direction '
                                'and mode specfied')
MODE_AND_ANCESTOR_SPECIFIED = ('<property> tag attribute "mode" cannot be '
                               'specifed with "ancestor"')


def IndexesXmlToIndexDefinitions(xml_str):
  """Converts a `<datastore-indexes>` XML string into an `IndexDefinitions` objects.

  Args:
    xml_str: A string containing a complete XML document where the root node is
      `<datastore-indexes>`.

  Returns:
    An `IndexDefinitions` object parsed out of the XML string.

  Raises:
    ValidationError: In case of malformed XML or illegal inputs.
  """
  parser = IndexesXmlParser()
  return parser.Parse(xml_str)


def IsAutoGenerated(xml_str):
  """Tests if the given `datastore-indexes.xml` string implies auto-generation."""
  try:
    xml_root = ElementTree.fromstring(xml_str)
    return (xml_root.tag == 'datastore-indexes' and
            _BooleanAttribute(xml_root.attrib.get('autoGenerate', 'false')))
  except ElementTree.ParseError:
    return False


class IndexesXmlParser(object):
  """Provides logic for walking down XML tree and pulling data."""

  def Parse(self, xml_str):
    """Parses XML string and returns object representation of relevant info.

    Args:
      xml_str: The XML string.
    Returns:
      An `IndexDefinitions` object containing the result of parsing the XML.
    Raises:
      ValidationError: In case of malformed XML or illegal inputs.
    """

    try:
      self.indexes = []
      self.errors = []
      xml_root = ElementTree.fromstring(xml_str)
      if xml_root.tag != 'datastore-indexes':
        raise ValidationError('Root tag must be <datastore-indexes>')

      for child in xml_root:
        self.ProcessIndexNode(child)

      if self.errors:
        raise ValidationError('\n'.join(self.errors))

      return IndexDefinitions(indexes=self.indexes)
    except ElementTree.ParseError as e:
      raise ValidationError('Bad input -- not valid XML: %s' % e)

  def ProcessIndexNode(self, node):
    """Processes XML `<datastore-index>` nodes into `Index` objects.

    The following information is parsed out:

    - `kind`: Specifies the kind of entities to index.
    - `ancestor`: `True` if the index supports queries that filter by
      `ancestor-key` to constraint results to a single entity group.
    - `property`: Represents the entity properties to index, with a name
       and direction attribute.

    Args:
      node: `<datastore-index>` XML node in `datastore-indexes.xml`.
    """
    if node.tag != 'datastore-index':
      self.errors.append('Unrecognized node: <%s>' % node.tag)
      return

    index = Index()
    index.kind = node.attrib.get('kind')
    if not index.kind:
      self.errors.append(MISSING_KIND)
    ancestor = node.attrib.get('ancestor', 'false')
    index.ancestor = _BooleanAttribute(ancestor)
    if index.ancestor is None:
      self.errors.append(
          'Value for ancestor should be true or false, not "%s"' % ancestor)
    properties = []
    property_nodes = [n for n in node if n.tag == 'property']


    has_geospatial = any(
        property_node.attrib.get('mode') == 'geospatial'
        for property_node in property_nodes)

    for property_node in property_nodes:
      name = property_node.attrib.get('name', '')
      if not name:
        self.errors.append(NAME_MISSING % index.kind)
        continue

      direction = property_node.attrib.get('direction')
      mode = property_node.attrib.get('mode')
      if mode:
        if index.ancestor:
          self.errors.append(MODE_AND_ANCESTOR_SPECIFIED)
          continue
        if mode != 'geospatial':
          self.errors.append(BAD_MODE % mode)
          continue
        if direction:
          self.errors.append(MODE_AND_DIRECTION_SPECIFIED)
          continue
      else:
        if not direction:


          if not has_geospatial:
            direction = 'asc'
        elif direction not in ('asc', 'desc'):
          self.errors.append(BAD_DIRECTION % direction)
          continue
      properties.append(Property(name=name, direction=direction, mode=mode))
    index.properties = properties
    self.indexes.append(index)


def _BooleanAttribute(value):
  """Parses the given attribute value as a Boolean value.

  This follows the specification here:
  http://www.w3.org/TR/2012/REC-xmlschema11-2-20120405/datatypes.html#boolean

  Args:
    value: The value to parse.

  Returns:
    `True` if the value parses as true, `False` if it parses as false, `None`
    if it parses as neither.
  """
  if value in ['true', '1']:
    return True
  elif value in ['false', '0']:
    return False
  else:
    return None



#!/usr/bin/env python
#
# Copyright 2007 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
"""Allows API stubs to access request and system state when handling calls.

Certain API stubs require access to information about the request that triggered
the API call (e.g., `user_service_stub` needs to know the host name of the
request to generate continuation URLs) or system state (e.g., `modules_stub`).

Other stubs (e.g., `taskqueue_stub`, `channel_stub`) need to be able to dispatch
requests within the system.

An instance of a `RequestInfo` subclass is passed to stubs that require these
capabilities.
"""

import collections
import logging

from google.appengine.runtime import context
from six.moves import urllib


class Error(Exception):
  pass


class ModuleDoesNotExistError(Error):
  """The provided module does not exist."""


class VersionDoesNotExistError(Error):
  """The provided version does not exist."""


class InvalidInstanceIdError(Error):
  """The provided instance ID is invalid."""


class NotSupportedWithAutoScalingError(Error):
  """The requested operation is not supported for auto-scaling modules."""


class VersionAlreadyStartedError(Error):
  """The version is already started."""


class VersionAlreadyStoppedError(Error):
  """The version is already stopped."""


class BackgroundThreadLimitReachedError(Error):
  """The instance is at its background thread capacity."""


ResponseTuple = collections.namedtuple('ResponseTuple',
                                       ['status', 'headers', 'content'])


class Dispatcher(object):
  """Provides information about and dispatches requests to modules."""

  def get_module_names(self):
    """Returns a list of module names."""
    raise NotImplementedError()

  def get_versions(self, module):
    """Returns a list of versions for a module.

    Args:
      module: A str containing the name of the module.

    Returns:
      A list of str containing the versions for the specified module.

    Raises:
      ModuleDoesNotExistError: The module does not exist.
    """
    raise NotImplementedError()

  def get_default_version(self, module):
    """Returns the default version for a module.

    Args:
      module: A str containing the name of the module.

    Returns:
      A str containing the default version for the specified module.

    Raises:
      ModuleDoesNotExistError: The module does not exist.
    """
    raise NotImplementedError()

  def get_hostname(self, module, version, instance=None):
    """Returns the hostname for a `(module, version, instance)` tuple.

    If instance is set, this will return a hostname for that particular
    instances. Otherwise, it will return the hostname for load-balancing.

    Args:
      module: A str containing the name of the module.
      version: A str containing the version.
      instance: An optional str containing the instance `ID`.

    Returns:
      A str containing the hostname.

    Raises:
      ModuleDoesNotExistError: The module does not exist.
      VersionDoesNotExistError: The version does not exist.
      InvalidInstanceIdError: The instance `ID` is not valid for the
          module/version or the module/version uses automatic scaling.
    """
    raise NotImplementedError()

  def set_num_instances(self, module, version, instances):
    """Sets the number of instances to run for a version of a module.

    Args:
      module: A str containing the name of the module.
      version: A str containing the version.
      instances: An int containing the number of instances to run.

    Raises:
      ModuleDoesNotExistError: The module does not exist.
      VersionDoesNotExistError: The version does not exist.
      NotSupportedWithAutoScalingError: The provided module/version uses
          automatic scaling.
    """
    raise NotImplementedError()

  def get_num_instances(self, module, version):
    """Gets the number of instances running for a version of a module.

    Args:
      module: A str containing the name of the module.
      version: A str containing the version.

    Raises:
      ModuleDoesNotExistError: The module does not exist.
      VersionDoesNotExistError: The version does not exist.
      NotSupportedWithAutoScalingError: The provided module/version uses
          automatic scaling.
    """
    raise NotImplementedError()

  def start_version(self, module, version):
    """Starts a version.

    Args:
      module: A str containing the name of the module.
      version: A str containing the version.

    Raises:
      ModuleDoesNotExistError: The module does not exist.
      VersionDoesNotExistError: The version does not exist.
      NotSupportedWithAutoScalingError: The provided module/version uses
          automatic scaling.
    """
    raise NotImplementedError()

  def stop_version(self, module, version):
    """Stops a version.

    Args:
      module: A str containing the name of the module.
      version: A str containing the version.

    Raises:
      ModuleDoesNotExistError: The module does not exist.
      VersionDoesNotExistError: The version does not exist.
      NotSupportedWithAutoScalingError: The provided module/version uses
          automatic scaling.
    """
    raise NotImplementedError()

  def add_event(self, runnable, eta, service=None, event_id=None):
    """Add a callable to be run at the specified time.

    Args:
      runnable: A callable object to call at the specified time.
      eta: An int containing the time to run the event, in seconds since the
          epoch.
      service: A str containing the name of the service that owns this event.
          This should be set if `event_id` is set.
      event_id: A str containing the id of the event. If set, this can be passed
          to update_event to change the time at which the event should run.
    """
    raise NotImplementedError()

  def update_event(self, eta, service, event_id):
    """Update the eta of a scheduled event.

    Args:
      eta: An int containing the time to run the event, in seconds since the
          epoch.
      service: A str containing the name of the service that owns this event.
      event_id: A str containing the id of the event to update.
    """
    raise NotImplementedError()

  def add_request(self, method, relative_url, headers, body, source_ip,
                  module_name=None, version=None, instance_id=None):
    """Process an HTTP request.

    Args:
      method: A str containing the HTTP method of the request.
      relative_url: A str containing path and query string of the request.
      headers: A list of (key, value) tuples where key and value are both str.
      body: A str containing the request body.
      source_ip: The source ip address for the request.
      module_name: An optional str containing the module name to service this
          request. If unset, the request will be dispatched to the default
          module.
      version: An optional str containing the version to service this request.
          If unset, the request will be dispatched to the default version.
      instance_id: An optional str containing the `instance_id` of the instance
          to service this request. If unset, the request will be dispatched to
          according to the load-balancing for the module and version.

    Returns:
      A `ResponseTuple` containing the response information for the HTTP
      request.
    """
    raise NotImplementedError()

  def add_async_request(self, method, relative_url, headers, body, source_ip,
                        module_name=None, version=None, instance_id=None):
    """Dispatch an HTTP request asynchronously.

    Args:
      method: A str containing the HTTP method of the request.
      relative_url: A str containing path and query string of the request.
      headers: A list of (key, value) tuples where key and value are both str.
      body: A str containing the request body.
      source_ip: The source ip address for the request.
      module_name: An optional str containing the module name to service this
          request. If unset, the request will be dispatched to the default
          module.
      version: An optional str containing the version to service this request.
          If unset, the request will be dispatched to the default version.
      instance_id: An optional str containing the instance_id of the instance to
          service this request. If unset, the request will be dispatched to
          according to the load-balancing for the module and version.
    """
    raise NotImplementedError()

  def send_background_request(self, module_name, version, instance,
                              background_request_id):
    """Dispatch a background thread request.

    Args:
      module_name: A str containing the module name to service this
          request.
      version: A str containing the version to service this request.
      instance: The instance to service this request.
      background_request_id: A str containing the unique background thread
          request identifier.

    Raises:
      NotSupportedWithAutoScalingError: The provided module/version uses
          automatic scaling.
      BackgroundThreadLimitReachedError: The instance is at its background
          thread capacity.
    """
    raise NotImplementedError()




class _LocalFakeDispatcher(Dispatcher):
  """A fake Dispatcher implementation usable by tests."""

  def __init__(self,
               module_names=None,
               module_name_to_versions=None,
               module_name_to_default_versions=None,
               module_name_to_version_to_hostname=None):
    super(_LocalFakeDispatcher, self).__init__()
    if module_names is None:
      module_names = ['default']
    if module_name_to_versions is None:
      module_name_to_versions = {'default': ['1']}
    if module_name_to_default_versions is None:
      module_name_to_default_versions = {'default': '1'}
    if module_name_to_version_to_hostname is None:
      module_name_to_version_to_hostname = {'default': {'1': 'localhost:8080'}}
    self._module_names = module_names
    self._module_name_to_versions = module_name_to_versions
    self._module_name_to_default_versions = module_name_to_default_versions
    self._module_name_to_version_to_hostname = (
        module_name_to_version_to_hostname)

  def get_module_names(self):
    """Returns a list of module names."""
    return self._module_names

  def get_versions(self, module):
    """Returns a list of versions for a module.

    Args:
      module: A str containing the name of the module.

    Returns:
      A list of str containing the versions for the specified module.

    Raises:
      ModuleDoesNotExistError: The module does not exist.
    """
    if module not in self._module_name_to_versions:
      raise ModuleDoesNotExistError()
    return self._module_name_to_versions[module]

  def get_default_version(self, module):
    """Returns the default version for a module.

    Args:
      module: A str containing the name of the module.

    Returns:
      A str containing the default version for the specified module.

    Raises:
      ModuleDoesNotExistError: The module does not exist.
    """
    if module not in self._module_name_to_default_versions:
      raise ModuleDoesNotExistError()
    return self._module_name_to_default_versions[module]

  def get_hostname(self, module, version, instance=None):
    """Returns the hostname for a `(module, version, instance)` tuple.

    If instance is set, this will return a hostname for that particular
    instances. Otherwise, it will return the hostname for load-balancing.

    Args:
      module: A str containing the name of the module.
      version: A str containing the version.
      instance: An optional str containing the instance `ID`.

    Returns:
      A str containing the hostname.

    Raises:
      ModuleDoesNotExistError: The module does not exist.
      VersionDoesNotExistError: The version does not exist.
      InvalidInstanceIdError: The instance `ID` is not valid for the
          module/version or the module/version uses automatic scaling.
    """
    if module not in self._module_name_to_version_to_hostname:
      raise ModuleDoesNotExistError()
    if version not in self._module_name_to_version_to_hostname[module]:
      raise VersionDoesNotExistError()
    if instance:

      raise InvalidInstanceIdError()
    return self._module_name_to_version_to_hostname[module][version]

  def set_num_instances(self, module, version, instances):
    """Sets the number of instances to run for a version of a module.

    Args:
      module: A str containing the name of the module.
      version: A str containing the version.
      instances: An int containing the number of instances to run.

    Raises:
      ModuleDoesNotExistError: The module does not exist.
      VersionDoesNotExistError: The version does not exist.
      NotSupportedWithAutoScalingError: The provided module/version uses
          automatic scaling.
    """
    if module not in self._module_name_to_versions:
      raise ModuleDoesNotExistError()
    if version not in self._module_name_to_versions[module]:
      raise VersionDoesNotExistError()

    raise NotSupportedWithAutoScalingError()

  def get_num_instances(self, module, version):
    """Gets the number of instances running for a version of a module.

    Args:
      module: A str containing the name of the module.
      version: A str containing the version.

    Raises:
      ModuleDoesNotExistError: The module does not exist.
      VersionDoesNotExistError: The version does not exist.
      NotSupportedWithAutoScalingError: The provided module/version uses
          automatic scaling.
    """
    if module not in self._module_name_to_versions:
      raise ModuleDoesNotExistError()
    if version not in self._module_name_to_versions[module]:
      raise VersionDoesNotExistError()

    raise NotSupportedWithAutoScalingError()

  def start_version(self, module, version):
    """Starts a version.

    Args:
      module: A str containing the name of the module.
      version: A str containing the version.

    Raises:
      ModuleDoesNotExistError: The module does not exist.
      VersionDoesNotExistError: The version does not exist.
      NotSupportedWithAutoScalingError: The provided module/version uses
          automatic scaling.
    """
    if module not in self._module_name_to_versions:
      raise ModuleDoesNotExistError()
    if version not in self._module_name_to_versions[module]:
      raise VersionDoesNotExistError()

    raise NotSupportedWithAutoScalingError()

  def stop_version(self, module, version):
    """Stops a version.

    Args:
      module: A str containing the name of the module.
      version: A str containing the version.

    Raises:
      ModuleDoesNotExistError: The module does not exist.
      VersionDoesNotExistError: The version does not exist.
      NotSupportedWithAutoScalingError: The provided module/version uses
          automatic scaling.
    """
    if module not in self._module_name_to_versions:
      raise ModuleDoesNotExistError()
    if version not in self._module_name_to_versions[module]:
      raise VersionDoesNotExistError()

    raise NotSupportedWithAutoScalingError()

  def add_event(self, runnable, eta, service=None, event_id=None):
    """Add a callable to be run at the specified time.

    Args:
      runnable: A callable object to call at the specified time.
      eta: An int containing the time to run the event, in seconds since the
          epoch.
      service: A str containing the name of the service that owns this event.
          This should be set if event_id is set.
      event_id: A str containing the id of the event. If set, this can be passed
          to update_event to change the time at which the event should run.
    """
    logging.warning('Scheduled events are not supported with '
                    '_LocalFakeDispatcher')

  def update_event(self, eta, service, event_id):
    """Update the eta of a scheduled event.

    Args:
      eta: An int containing the time to run the event, in seconds since the
          epoch.
      service: A str containing the name of the service that owns this event.
      event_id: A str containing the id of the event to update.
    """
    logging.warning('Scheduled events are not supported with '
                    '_LocalFakeDispatcher')

  def add_request(self, method, relative_url, headers, body, source_ip,
                  module_name=None, version=None, instance_id=None):
    """Process an HTTP request.

    Args:
      method: A str containing the HTTP method of the request.
      relative_url: A str containing path and query string of the request.
      headers: A list of `(key, value)` tuples where key and value are both str.
      body: A str containing the request body.
      source_ip: The source ip address for the request.
      module_name: An optional str containing the module name to service this
          request. If unset, the request will be dispatched to the default
          module.
      version: An optional str containing the version to service this request.
          If unset, the request will be dispatched to the default version.
      instance_id: An optional str containing the instance_id of the instance to
          service this request. If unset, the request will be dispatched to
          according to the load-balancing for the module and version.

    Returns:
      A `ResponseTuple` containing the response information for the `HTTP`
      request.
    """
    logging.warning('Request dispatching is not supported with '
                    '_LocalFakeDispatcher')
    return ResponseTuple('501 Not Implemented', [], '')

  def add_async_request(self, method, relative_url, headers, body, source_ip,
                        module_name=None, version=None, instance_id=None):
    """Dispatch an HTTP request asynchronously.

    Args:
      method: A str containing the `HTTP` method of the request.
      relative_url: A str containing path and query string of the request.
      headers: A list of `(key, value)` tuples where key and value are both str.
      body: A str containing the request body.
      source_ip: The source ip address for the request.
      module_name: An optional str containing the module name to service this
          request. If unset, the request will be dispatched to the default
          module.
      version: An optional str containing the version to service this request.
          If unset, the request will be dispatched to the default version.
      instance_id: An optional str containing the instance_id of the instance to
          service this request. If unset, the request will be dispatched to
          according to the load-balancing for the module and version.
    """
    logging.warning('Request dispatching is not supported with '
                    '_LocalFakeDispatcher')

  def send_background_request(self, module_name, version, instance,
                              background_request_id):
    """Dispatch a background thread request.

    Args:
      module_name: A str containing the module name to service this
          request.
      version: A str containing the version to service this request.
      instance: The instance to service this request.
      background_request_id: A str containing the unique background thread
          request identifier.

    Raises:
      NotSupportedWithAutoScalingError: The provided module/version uses
          automatic scaling.
      BackgroundThreadLimitReachedError: The instance is at its background
          thread capacity.
    """
    logging.warning('Request dispatching is not supported with '
                    '_LocalFakeDispatcher')
    raise BackgroundThreadLimitReachedError()

_local_dispatcher = _LocalFakeDispatcher()


class RequestInfo(object):
  """Allows stubs to lookup state linked to the request making the API call."""

  def get_request_url(self, request_id):
    """Returns the URL the request e.g., `http://localhost:8080/foo?bar=baz`.

    Args:
      request_id: The string id of the request making the API call.

    Returns:
      The URL of the request as a string.
    """
    raise NotImplementedError()

  def get_request_environ(self, request_id):
    """Returns a dict containing the WSGI environ for the request."""
    raise NotImplementedError()

  def get_module(self, request_id):
    """Returns the name of the module serving this request.

    Args:
      request_id: The string id of the request making the API call.

    Returns:
      A str containing the module name.
    """
    raise NotImplementedError()

  def get_version(self, request_id):
    """Returns the version of the module serving this request.

    Args:
      request_id: The string id of the request making the API call.

    Returns:
      A str containing the version.
    """
    raise NotImplementedError()

  def get_instance(self, request_id):
    """Returns the instance serving this request.

    Args:
      request_id: The string id of the request making the API call.

    Returns:
      An opaque representation of the instance serving this request. It should
      only be passed to dispatcher methods expecting an instance.
    """
    raise NotImplementedError()

  def get_dispatcher(self):
    """Returns the Dispatcher.

    Returns:
      The Dispatcher instance.
    """
    raise NotImplementedError()


  def register_request_id(self, environ, request_id):
    """Registers a simulated HTTP request from the dev_appserver.

    Subclasses are not required to provide an implementation.

    Args:
      environ: An environ dict for the request as defined in PEP-333.
      request_id: The string id of the request making the API call.
    """


class _LocalRequestInfo(RequestInfo):
  """Lookup information about a request using environment variables."""

  def get_request_url(self, request_id):
    """Returns the URL the request e.g., `http://localhost:8080/foo?bar=baz`.

    Args:
      request_id: The string id of the request making the API call.

    Returns:
      The URL of the request as a string.
    """
    host = context.get('HTTP_HOST')
    if not host:
      host = context.get('SERVER_NAME')
      port = context.get('SERVER_PORT')
      if port != '80':
        host += ':' + port
    url = 'http://' + host
    url += urllib.parse.quote(context.get('PATH_INFO', '/'))
    if context.get('QUERY_STRING'):
      url += '?' + context.get('QUERY_STRING')
    return url

  def get_module(self, request_id):
    """Returns the name of the module serving this request.

    Args:
      request_id: The string id of the request making the API call.

    Returns:
      A str containing the module name.
    """
    return 'default'

  def get_version(self, request_id):
    """Returns the version of the module serving this request.

    Args:
      request_id: The string id of the request making the API call.

    Returns:
      A str containing the version.
    """
    return '1'

  def get_instance(self, request_id):
    """Returns the instance serving this request.

    Args:
      request_id: The string id of the request making the API call.

    Returns:
      An opaque representation of the instance serving this request. It should
      only be passed to dispatcher methods expecting an instance.
    """
    return object()

  def get_dispatcher(self):
    """Returns the Dispatcher.

    Returns:
      The Dispatcher instance.
    """
    return _local_dispatcher


_local_request_info = _LocalRequestInfo()



#!/usr/bin/env python
#
# Copyright 2007 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#



"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database


_sym_db = _symbol_database.Default()




DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n\'google/appengine/api/mail_service.proto\x12\x10google.appengine\"\xb4\x01\n\x10MailServiceError\"\x9f\x01\n\tErrorCode\x12\x06\n\x02OK\x10\x00\x12\x12\n\x0eINTERNAL_ERROR\x10\x01\x12\x0f\n\x0b\x42\x41\x44_REQUEST\x10\x02\x12\x17\n\x13UNAUTHORIZED_SENDER\x10\x03\x12\x1b\n\x17INVALID_ATTACHMENT_TYPE\x10\x04\x12\x17\n\x13INVALID_HEADER_NAME\x10\x05\x12\x16\n\x12INVALID_CONTENT_ID\x10\x06\"i\n\x0eMailAttachment\x12\x10\n\x08\x46ileName\x18\x01 \x02(\t\x12\x0c\n\x04\x44\x61ta\x18\x02 \x02(\x0c\x12\x11\n\tContentID\x18\x03 \x01(\t\x12$\n\x18\x44\x45PRECATED_ContentID_set\x18\r \x01(\x08\x42\x02\x18\x01\")\n\nMailHeader\x12\x0c\n\x04name\x18\x01 \x02(\t\x12\r\n\x05value\x18\x02 \x02(\t\"\x81\x02\n\x0bMailMessage\x12\x0e\n\x06Sender\x18\x01 \x02(\t\x12\x0f\n\x07ReplyTo\x18\x02 \x01(\t\x12\n\n\x02To\x18\x03 \x03(\t\x12\n\n\x02\x43\x63\x18\x04 \x03(\t\x12\x0b\n\x03\x42\x63\x63\x18\x05 \x03(\t\x12\x0f\n\x07Subject\x18\x06 \x02(\t\x12\x10\n\x08TextBody\x18\x07 \x01(\t\x12\x10\n\x08HtmlBody\x18\x08 \x01(\t\x12\x13\n\x0b\x41mpHtmlBody\x18\x0b \x01(\t\x12\x34\n\nAttachment\x18\t \x03(\x0b\x32 .google.appengine.MailAttachment\x12,\n\x06Header\x18\n \x03(\x0b\x32\x1c.google.appengine.MailHeaderB.\n\x1d\x63om.google.appengine.api.mailB\rMailServicePb')



_MAILSERVICEERROR = DESCRIPTOR.message_types_by_name['MailServiceError']
_MAILATTACHMENT = DESCRIPTOR.message_types_by_name['MailAttachment']
_MAILHEADER = DESCRIPTOR.message_types_by_name['MailHeader']
_MAILMESSAGE = DESCRIPTOR.message_types_by_name['MailMessage']
_MAILSERVICEERROR_ERRORCODE = _MAILSERVICEERROR.enum_types_by_name['ErrorCode']
MailServiceError = _reflection.GeneratedProtocolMessageType('MailServiceError', (_message.Message,), {
  'DESCRIPTOR' : _MAILSERVICEERROR,
  '__module__' : 'google.appengine.api.mail_service_pb2'

  })
_sym_db.RegisterMessage(MailServiceError)

MailAttachment = _reflection.GeneratedProtocolMessageType('MailAttachment', (_message.Message,), {
  'DESCRIPTOR' : _MAILATTACHMENT,
  '__module__' : 'google.appengine.api.mail_service_pb2'

  })
_sym_db.RegisterMessage(MailAttachment)

MailHeader = _reflection.GeneratedProtocolMessageType('MailHeader', (_message.Message,), {
  'DESCRIPTOR' : _MAILHEADER,
  '__module__' : 'google.appengine.api.mail_service_pb2'

  })
_sym_db.RegisterMessage(MailHeader)

MailMessage = _reflection.GeneratedProtocolMessageType('MailMessage', (_message.Message,), {
  'DESCRIPTOR' : _MAILMESSAGE,
  '__module__' : 'google.appengine.api.mail_service_pb2'

  })
_sym_db.RegisterMessage(MailMessage)

if _descriptor._USE_C_DESCRIPTORS == False:

  DESCRIPTOR._options = None
  DESCRIPTOR._serialized_options = b'\n\035com.google.appengine.api.mailB\rMailServicePb'
  _MAILATTACHMENT.fields_by_name['DEPRECATED_ContentID_set']._options = None
  _MAILATTACHMENT.fields_by_name['DEPRECATED_ContentID_set']._serialized_options = b'\030\001'
  _MAILSERVICEERROR._serialized_start=62
  _MAILSERVICEERROR._serialized_end=242
  _MAILSERVICEERROR_ERRORCODE._serialized_start=83
  _MAILSERVICEERROR_ERRORCODE._serialized_end=242
  _MAILATTACHMENT._serialized_start=244
  _MAILATTACHMENT._serialized_end=349
  _MAILHEADER._serialized_start=351
  _MAILHEADER._serialized_end=392
  _MAILMESSAGE._serialized_start=395
  _MAILMESSAGE._serialized_end=652




#!/usr/bin/env python
#
# Copyright 2007 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#


"""Allows applications to identify API outages and scheduled downtime.

   Example:

   ```python
    def StoreUploadedProfileImage(self):
      uploaded_image = self.request.get('img')
      # If the images API is unavailable, we'll just skip the resize.
      if CapabilitySet('images').is_enabled():
        uploaded_image = images.resize(uploaded_image, 64, 64)
      store(uploaded_image)

    def RenderHTMLForm(self):
      datastore_readonly = CapabilitySet('datastore_v3', capabilities=['write'])
      if datastore_readonly.is_enabled():
        # ...render form normally...
      else:
        # self.response.out('<p>Not accepting submissions right now: %s</p>' %
                            datastore_readonly.admin_message())
        # ...render form with form elements disabled...
     ```

Individual API wrapper modules should expose `CapabilitySet` objects
for users rather than relying on users to create them.  They can
also create convenience methods that delegate to the relevant `CapabilitySet`;
for example, `db.IsReadOnly()`.
"""










import warnings

from google.appengine.api import apiproxy_stub_map
from google.appengine.api.capabilities import capability_service_pb2
from google.appengine.base import capabilities_pb2


IsEnabledRequest = capability_service_pb2.IsEnabledRequest
IsEnabledResponse = capability_service_pb2.IsEnabledResponse
CapabilityConfig = capabilities_pb2.CapabilityConfig


class UnknownCapabilityError(Exception):
  """An unknown capability was requested."""


class CapabilitySet(object):
  """Encapsulates one or more capabilities.

  Capabilities can either be named explicitly, or inferred from the list of
  methods provided. If no capabilities or methods are provided, this will check
  whether the entire package is enabled.
  """

  def __init__(self, package, capabilities=None, methods=None,
               stub_map=apiproxy_stub_map):
    """Constructor.

    Args:
      capabilities: List of strings
      methods: List of strings
    """
    if capabilities is None:
      capabilities = []
    if methods is None:
      methods = []
    self._package = package
    self._capabilities = ['*'] + capabilities
    self._methods = methods
    self._stub_map = stub_map

  def is_enabled(self):
    """Tests whether the capabilities are currently enabled.

    Returns:
      `True` if API calls that require these capabilities will succeed.

    Raises:
      UnknownCapabilityError: If a specified capability was not recognized.
    """
    config = self._get_status()
    return config.summary_status in (IsEnabledResponse.DEFAULT,
                                     IsEnabledResponse.ENABLED,
                                     IsEnabledResponse.SCHEDULED_FUTURE,
                                     IsEnabledResponse.SCHEDULED_NOW)

  def will_remain_enabled_for(self, time=60):
    """Returns whether a capability will remain enabled.

    DEPRECATED: Use `is_enabled()` instead. This method was never fully
    implemented.

    Args:
      time: Number of seconds in the future to look when checking for scheduled
          downtime.

    Returns:
      `True` if there is no scheduled downtime for the specified capability
      within the amount of time specified.

    Raises:
      UnknownCapabilityError: If a specified capability was not recognized.
    """
    warnings.warn('will_remain_enabled_for() is deprecated: '
                  'use is_enabled instead.',
                  DeprecationWarning,
                  stacklevel=2)
    config = self._get_status()

    status = config.summary_status

    if status in (IsEnabledResponse.DEFAULT, IsEnabledResponse.ENABLED):
      return True
    elif status == IsEnabledResponse.SCHEDULED_NOW:
      return False
    elif status == IsEnabledResponse.SCHEDULED_FUTURE:
      if config.HasField('time_until_scheduled'):
        return config.time_until_scheduled >= time
      else:

        return True
    elif status == IsEnabledResponse.DISABLED:
      return False
    else:


      return False

  def admin_message(self):
    """Retrieves any administrator notice messages for these capabilities.

    Returns:
      A string containing one or more administrator messages, or an empty
      string.

    Raises:
      UnknownCapabilityError: If a specified capability was not recognized.
    """
    message_list = []
    for config in self._get_status().config:
      message = config.admin_message
      if message and message not in message_list:
        message_list.append(message)
    return '  '.join(message_list)

  def _get_status(self):
    """Gets the `IsEnabledResponse` for the capabilities listed.

    Returns:
      `IsEnabledResponse` for the specified capabilities.

    Raises:
      UnknownCapabilityError: If an unknown capability was requested.
    """
    req = IsEnabledRequest()
    req.package = self._package
    for capability in self._capabilities:
      req.capability.append(capability)
    for method in self._methods:
      req.call.append(method)

    resp = capability_service_pb2.IsEnabledResponse()
    if (self._package == 'datastore_v3' and 'write' in self._capabilities or
        self._methods != 0):
      self._stub_map.MakeSyncCall('capability_service', 'IsEnabled', req, resp)
    else:
      resp = IsEnabledResponse()
      resp.set_summary_status(IsEnabledResponse.ENABLED)
      for capability in self._capabilities:
        resp.add_config(
            package=self._package,
            capability=capability,
            status=CapabilityConfig.ENABLED)

    return resp




#!/usr/bin/env python
#
# Copyright 2007 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#



"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database


_sym_db = _symbol_database.Default()




DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n\'google/appengine/datastore/action.proto\x12\x13storage_onestore_v3\"\x93\x01\n\x06\x41\x63tion\x12>\n\x10transaction_data\x18\x02 \x01(\x0b\x32$.storage_onestore_v3.TransactionData\x12;\n\x0f\x61\x63tion_rpc_info\x18\x03 \x01(\x0b\x32\".storage_onestore_v3.ActionRpcInfo\x12\x0c\n\x04uuid\x18\x04 \x01(\t\"F\n\x0fTransactionData\x12\x0e\n\x06handle\x18\x01 \x01(\x06\x12\x0e\n\x06\x61pp_id\x18\x02 \x01(\t\x12\x13\n\x0b\x64\x61tabase_id\x18\x03 \x01(\t\"Z\n\rActionRpcInfo\x12\x1f\n\x17\x65nqueuing_rpc_global_id\x18\x01 \x01(\x03\x12(\n enqueuing_rpc_start_timestamp_us\x18\x02 \x01(\x03\x42\x45\n\x1e\x63om.google.storage.onestore.v3B\x0eOnestoreActionZ\x13storage_onestore_v3')



_ACTION = DESCRIPTOR.message_types_by_name['Action']
_TRANSACTIONDATA = DESCRIPTOR.message_types_by_name['TransactionData']
_ACTIONRPCINFO = DESCRIPTOR.message_types_by_name['ActionRpcInfo']
Action = _reflection.GeneratedProtocolMessageType('Action', (_message.Message,), {
  'DESCRIPTOR' : _ACTION,
  '__module__' : 'google.appengine.datastore.action_pb2'

  })
_sym_db.RegisterMessage(Action)

TransactionData = _reflection.GeneratedProtocolMessageType('TransactionData', (_message.Message,), {
  'DESCRIPTOR' : _TRANSACTIONDATA,
  '__module__' : 'google.appengine.datastore.action_pb2'

  })
_sym_db.RegisterMessage(TransactionData)

ActionRpcInfo = _reflection.GeneratedProtocolMessageType('ActionRpcInfo', (_message.Message,), {
  'DESCRIPTOR' : _ACTIONRPCINFO,
  '__module__' : 'google.appengine.datastore.action_pb2'

  })
_sym_db.RegisterMessage(ActionRpcInfo)

if _descriptor._USE_C_DESCRIPTORS == False:

  DESCRIPTOR._options = None
  DESCRIPTOR._serialized_options = b'\n\036com.google.storage.onestore.v3B\016OnestoreActionZ\023storage_onestore_v3'
  _ACTION._serialized_start=65
  _ACTION._serialized_end=212
  _TRANSACTIONDATA._serialized_start=214
  _TRANSACTIONDATA._serialized_end=284
  _ACTIONRPCINFO._serialized_start=286
  _ACTIONRPCINFO._serialized_end=376




#!/usr/bin/env python
#
# Copyright 2007 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#


"""Stub version of the blob-related parts of the images API."""






import logging

from google.appengine.api import datastore


BLOB_SERVING_URL_KIND = '__BlobServingUrl__'


class ImagesBlobStub(object):
  """Stub version of the blob-related parts of the images API."""

  def __init__(self, host_prefix):
    """Stub implementation of blob-related parts of the images API.

    Args:
      host_prefix: string - The URL prefix (protocol://host:port) to prepend to
        image URLs on a call to GetUrlBase.
    """
    self._host_prefix = host_prefix

  def GetUrlBase(self, request, response):
    """Trivial implementation of an API call.

    Args:
      request: ImagesGetUrlBaseRequest - Contains a blobkey to an image.
      response: ImagesGetUrlBaseResponse - Contains a URL to serve the image.
    """
    if request.create_secure_url:
      logging.info('Secure URLs will not be created using the development '
                   'application server.')

    entity_info = datastore.Entity(
        BLOB_SERVING_URL_KIND, name=request.blob_key, namespace='')
    entity_info['blob_key'] = request.blob_key
    datastore.Put(entity_info)

    response.url = '%s/_ah/img/%s' % (self._host_prefix, request.blob_key)

  def DeleteUrlBase(self, request, unused_response):
    """Trivial implementation of an API call.

    Args:
      request: ImagesDeleteUrlBaseRequest - Contains a blobkey to an image.
      unused_response: ImagesDeleteUrlBaseResponse - Unused.
    """
    key = datastore.Key.from_path(
        BLOB_SERVING_URL_KIND, request.blob_key, namespace='')
    datastore.Delete(key)



#!/usr/bin/env python
#
# Copyright 2007 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#


"""Base class for implementing RPC of API proxy stubs."""








from concurrent import futures
import sys
import contextvars








_MAX_CONCURRENT_API_CALLS = 100

_THREAD_POOL = futures.ThreadPoolExecutor(_MAX_CONCURRENT_API_CALLS)


class RPC(object):
  """Base class for implementing RPC of API proxy stubs.

  Constructor for the RPC object.

  All arguments are optional, and simply set members on the class.
  These data members will be overriden by values passed to `MakeCall`.
  """

  IDLE = 0
  RUNNING = 1
  FINISHING = 2

  def __init__(self, package=None, call=None, request=None, response=None,
               callback=None, deadline=None, stub=None):
    """Constructor.

    Args:
      package: `string`. The package for the call.
      call: `string`. The call within the package.
      request: `ProtocolMessage` instance. Appropriate for the arguments.
      response: `ProtocolMessage` instance. Appropriate for the response.
      callback: `callable`. Called when call is complete.
      deadline: `double`. Specifies the deadline for this call as the number
        of seconds from the current time. Ignored if non-positive.
      stub: `APIProxyStub` instance. Used in default `_WaitImpl` to do real
        call.
    """
    self._exception = None
    self._state = RPC.IDLE

    self.package = package
    self.call = call
    self.request = request
    self.future = None
    self.response = response
    self.callback = callback
    self.deadline = deadline
    self.stub = stub
    self.cpu_usage_mcycles = 0

  def Clone(self):
    """Make a shallow copy of this instances attributes, excluding methods.

    This is usually used when an RPC has been specified with some configuration
    options and is being used as a template for multiple RPCs outside of a
    developer's easy control.

    Returns:
      A clone of this RPC.
    """
    if self.state != RPC.IDLE:
      raise AssertionError('Cannot clone a call already in progress')

    clone = self.__class__()
    for k, v in self.__dict__.items():
      setattr(clone, k, v)
    return clone

  def MakeCall(self, package=None, call=None, request=None, response=None,
               callback=None, deadline=None):
    """Makes an asynchronous (i.e., non-blocking) API call within the specified package for the specified call method.

    It will call the `_MakeRealCall` to do the real job.

    Args:
      package: `string`. The package for the call.
      call: `string`. The call within the package.
      request: `ProtocolMessage` instance. Appropriate for the arguments.
      response: `ProtocolMessage` instance. Appropriate for the response.
      callback: `callable`. Called when call is complete.
      deadline: `double`. Specifies the deadline for this call as the number
        of seconds from the current time. Ignored if non-positive.

    Raises:
      `TypeError` or `AssertionError` if an argument is of an invalid type.
      `AssertionError` or `RuntimeError` is an RPC is already in use.
    """
    self.callback = callback or self.callback
    self.package = package or self.package
    self.call = call or self.call
    self.request = request or self.request
    self.response = response or self.response
    self.deadline = deadline or self.deadline
    self.future = None

    assert self._state == RPC.IDLE, ('RPC for %s.%s has already been started' %
                                     (self.package, self.call))
    assert self.callback is None or callable(self.callback)
    self._MakeCallImpl()

  def Wait(self):
    """Waits on the API call associated with this RPC."""
    rpc_completed = self._WaitImpl()

    assert rpc_completed, ('RPC for %s.%s was not completed, and no other '
                           'exception was raised ' % (self.package, self.call))

  def CheckSuccess(self):
    """If there was an exception, raise it now.

    Raises:
      Exception of the API call or the `callback`, if any.
    """
    if self.exception:
      raise self.exception

  @property
  def exception(self):
    return self._exception

  @property
  def state(self):
    return self._state

  def _MakeCallImpl(self):
    """Makes an asynchronous API call.

    For this to work the following must be set:
      `self.package`: the API package name;
      `self.call`: the name of the API call/method to invoke;
      `self.request`: the API request body as a serialized protocol buffer.

    The actual API call is made via a thread pool. The thread pool restricts the
    number of concurrent requests to `MAX_CONCURRENT_API_CALLS`, so this method
    will block if that limit is exceeded, until other asynchronous calls
    resolve.

    If the main thread holds the import lock, waiting on thread work can cause
    a deadlock:
    https://docs.python.org/2/library/threading.html#importing-in-threaded-code

    Therefore, we try to detect this error case and fall back to sync calls.
    """
    assert self._state == RPC.IDLE, self._state

    self._state = RPC.RUNNING


    run_async = getattr(self.stub.__class__, 'THREADSAFE', False)

    ctx = contextvars.copy_context()

    def ExecInContext():
      ctx.run(self._SendRequestAndFinish)

    if run_async:
      self.future = _THREAD_POOL.submit(ExecInContext)
    else:
      self.future = None

  def _WaitImpl(self):
    """Waits for this and only this RPC to complete."""

    if self.future:
      futures.wait([self.future])
      return True


    self._SendRequestAndFinish(reraise_callback_exception=True)
    return True

  def _SendRequest(self):
    self.stub.MakeSyncCall(self.package, self.call, self.request, self.response)

  def _CaptureTrace(self, f):
    """Runs `f()` and captures exception information if raised."""
    try:
      f()
    except Exception as exc:


      self._exception = exc

  def _CaptureTraceAndReraise(self, f):
    """A variant of `_CaptureTrace()` used in the synchronous fallback path."""
    try:
      f()
    except:
      _, exc, _ = sys.exc_info()
      self._exception = exc


      self._exception._appengine_apiproxy_rpc = self
      raise

  def _SendRequestAndFinish(self, reraise_callback_exception=False):
    try:
      self._CaptureTrace(self._SendRequest)
    finally:
      self._state = RPC.FINISHING
      if self.callback:
        if reraise_callback_exception:
          self._CaptureTraceAndReraise(self.callback)
        else:
          self._CaptureTrace(self.callback)



#!/usr/bin/env python
#
# Copyright 2007 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#


"""Models and helper functions for access to app's datastore metadata.

These entities cannot be created by users, but are created as results of
`__namespace__`, `__kind__` and `__property__` metadata queries such as

  # Find all namespaces
  q = db.GqlQuery("SELECT * FROM __namespace__")
  for p in q.run():
    print "namespace:", repr(p.namespace_name)

or

  # Find all properties of A whose name starts with a lower-case letter
  q = db.GqlQuery("SELECT __key__ from __property__ " +
                  "WHERE __key__ >= :1 AND __key__ < :2",
                  Property.key_for_property("A", "a"),
                  Property.key_for_property("A", chr(ord("z") + 1)))
  for p in q.run():
    print "%s: %s" % (Property.key_to_kind(p), Property.key_to_property(p))

or, using `Query` objects

  # Find all kinds >= "a"
  q = metadata.Kind().all()
  q.filter("__key__ >=", metadata.Kind.key_for_kind("a"))
  for p in q.run():
    print "kind:", repr(p.kind_name)
"""









from google.appengine.api import datastore_types
from google.appengine.ext import db


class BaseMetadata(db.Model):
  """Base class for all metadata models."""


  KIND_NAME = '__BaseMetadata__'

  @classmethod
  def kind(cls):
    """Kind name override."""
    return cls.KIND_NAME


class Namespace(BaseMetadata):
  """Model for `__namespace__` metadata query results."""

  KIND_NAME = '__namespace__'
  EMPTY_NAMESPACE_ID = datastore_types._EMPTY_NAMESPACE_ID

  @property
  def namespace_name(self):
    """Return the namespace name specified by this entity's key."""
    return self.key_to_namespace(self.key())

  @classmethod
  def key_for_namespace(cls, namespace):
    """Return the `__namespace__` key for namespace.

    Args:
      namespace: namespace whose key is requested.

    Returns:
      The key for namespace.
    """
    if namespace:
      return db.Key.from_path(cls.KIND_NAME, namespace)
    else:
      return db.Key.from_path(cls.KIND_NAME, cls.EMPTY_NAMESPACE_ID)

  @classmethod
  def key_to_namespace(cls, key):
    """Return the namespace specified by a given `__namespace__` key.

    Args:
      key: key whose name is requested.

    Returns:
      The namespace specified by key.
    """
    return key.name() or ''


class Kind(BaseMetadata):
  """Model for `__kind__` metadata query results."""

  KIND_NAME = '__kind__'

  @property
  def kind_name(self):
    """Return the kind name specified by this entity's key."""
    return self.key_to_kind(self.key())

  @classmethod
  def key_for_kind(cls, kind):
    """Return the `__kind__` key for kind.

    Args:
      kind: kind whose key is requested.

    Returns:
      The key for kind.
    """
    return db.Key.from_path(cls.KIND_NAME, kind)

  @classmethod
  def key_to_kind(cls, key):
    """Return the kind specified by a given `__kind__` key.

    Args:
      key: key whose name is requested.

    Returns:
      The kind specified by key.
    """
    return key.name()


class Property(BaseMetadata):
  """Model for `__property__` metadata query results."""

  KIND_NAME = '__property__'

  @property
  def property_name(self):
    """Return the property name specified by this entity's key."""
    return self.key_to_property(self.key())

  @property
  def kind_name(self):
    """Return the kind name specified by this entity's key."""
    return self.key_to_kind(self.key())

  property_representation = db.StringListProperty()

  @classmethod
  def key_for_kind(cls, kind):
    """Return the `__property__` key for kind.

    Args:
      kind: kind whose key is requested.

    Returns:
      The parent key for `__property__` keys of kind.
    """
    return db.Key.from_path(Kind.KIND_NAME, kind)

  @classmethod
  def key_for_property(cls, kind, property_):
    """Return the `__property__` key for property of kind.

    Args:
      kind: kind whose key is requested.
      property_: property whose key is requested.

    Returns:
      The key for property of kind.
    """
    return db.Key.from_path(Kind.KIND_NAME, kind, Property.KIND_NAME, property_)

  @classmethod
  def key_to_kind(cls, key):
    """Return the kind specified by a given `__property__` key.

    Args:
      key: key whose kind name is requested.

    Returns:
      The kind specified by key.
    """
    if key.kind() == Kind.KIND_NAME:
      return key.name()
    else:
      return key.parent().name()

  @classmethod
  def key_to_property(cls, key):
    """Return the property specified by a given `__property__` key.

    Args:
      key: key whose property name is requested.

    Returns:
      property specified by key, or `None` if the key specified only a kind.
    """
    if key.kind() == Kind.KIND_NAME:
      return None
    else:
      return key.name()


class EntityGroup(BaseMetadata):
  """Model for `__entity_group__` metadata (available in HR datastore only).

  This metadata contains a numeric `__version__` property that is guaranteed
  to increase on every change to the entity group. The version may increase
  even in the absence of user-visible changes to the entity group. The
  `__entity_group__` entity may not exist `None` if the entity group was never
  written to.
  """

  KIND_NAME = '__entity_group__'
  ID = 1

  version = db.IntegerProperty(name='__version__')

  @classmethod
  def key_for_entity(cls, entity_or_key):
    """Return the metadata key for the entity group containing `entity_or_key`.

    Use this key to `get()` the `__entity_group__` metadata entity for the
    entity group containing `entity_or_key`.

    Args:
      entity_or_key: a key or entity whose `__entity_group__` key you want.

    Returns:
      The `__entity_group__` key for the entity group containing
      `entity_or_key`.
    """
    if isinstance(entity_or_key, db.Model):
      key = entity_or_key.key()
    else:
      key = entity_or_key
    while key.parent():
      key = key.parent()
    return db.Key.from_path(cls.KIND_NAME, cls.ID, parent=key)


def get_namespaces(start=None, end=None):
  """Return all namespaces in the specified range.

  Args:
    start: only return namespaces >= start if start is not `None`.
    end: only return namespaces < end if end is not `None`.

  Returns:
    A list of namespace names between the (optional) start and end values.
  """
  q = Namespace.all()
  if start is not None:
    q.filter('__key__ >=', Namespace.key_for_namespace(start))
  if end is not None:
    q.filter('__key__ <', Namespace.key_for_namespace(end))

  return [x.namespace_name for x in q.run()]


def get_kinds(start=None, end=None):
  """Return all kinds in the specified range.

  Args:
    start: only return kinds >= start if start is not `None`.
    end: only return kinds < end if end is not `None`.

  Returns:
    A list of kind names between the (optional) start and end values.
  """
  q = Kind.all()
  if start:
    q.filter('__key__ >=', Kind.key_for_kind(start))
  if end is not None:
    if end == '':
      return []
    q.filter('__key__ <', Kind.key_for_kind(end))

  return [x.kind_name for x in q.run()]


def get_properties_of_kind(kind, start=None, end=None):
  """Return all properties of kind in the specified range.

  NOTE: This function does not return unindexed properties.

  Args:
    kind: name of kind whose properties you want.
    start: only return properties >= start if start is not `None`.
    end: only return properties < end if end is not `None`.

  Returns:
    A list of property names of kind between the (optional) start and end
    values.
  """
  q = Property.all(keys_only=True)
  q.ancestor(Property.key_for_kind(kind))
  if start:
    q.filter('__key__ >=', Property.key_for_property(kind, start))
  if end is not None:
    if end == '':
      return []
    q.filter('__key__ <', Property.key_for_property(kind, end))

  return [Property.key_to_property(x) for x in q.run()]


def get_representations_of_kind(kind, start=None, end=None):
  """Return all representations of properties of kind in the specified range.

  NOTE: This function does not return unindexed properties.

  Args:
    kind: name of kind whose properties you want.
    start: only return properties >= start if start is not `None`.
    end: only return properties < end if end is not `None`.

  Returns:
    A dictionary mapping property names to its list of representations.
  """
  q = Property.all()
  q.ancestor(Property.key_for_kind(kind))
  if start:
    q.filter('__key__ >=', Property.key_for_property(kind, start))
  if end is not None:
    if end == '':
      return {}
    q.filter('__key__ <', Property.key_for_property(kind, end))

  result = {}
  for property_ in q.run():
    result[property_.property_name] = property_.property_representation

  return result


def get_entity_group_version(entity_or_key):
  """Returns the version of the entity group containing `entity_or_key`.

  Args:
    entity_or_key: a key or entity whose version you want.
  Returns: The version of the entity group containing `entity_or_key`. This
    version is guaranteed to increase on every change to the entity group. The
    version may increase even in the absence of user-visible changes to the
    entity group. May return `None` if the entity group was never written to. On
    non-HR datatores, this function returns `None`.
  """

  eg = db.get(EntityGroup.key_for_entity(entity_or_key))
  if eg:
    return eg.version
  else:
    return None



#!/usr/bin/env python
#
# Copyright 2007 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#



"""Trivial implementation of the `UserService`."""





import logging
from six.moves import urllib
import six.moves.urllib.parse
from google.appengine.api import apiproxy_stub
from google.appengine.api.oauth import oauth_api
from google.appengine.api import user_service_pb2
from google.appengine.runtime import apiproxy_errors
from google.appengine.runtime.context import ctx_test_util

_DEFAULT_LOGIN_URL = 'https://www.google.com/accounts/Login?continue=%s'
_DEFAULT_LOGOUT_URL = 'https://www.google.com/accounts/Logout?continue=%s'
_DEFAULT_AUTH_DOMAIN = 'gmail.com'




_OAUTH_CONSUMER_KEY = 'example.com'
_OAUTH_EMAIL = 'example@example.com'
_OAUTH_USER_ID = '0'
_OAUTH_AUTH_DOMAIN = _DEFAULT_AUTH_DOMAIN
_OAUTH_CLIENT_ID = '123456789.apps.googleusercontent.com'

_OPENID_DEPRECATION_WARNING = (
    'Open ID 2.0 support in the App Engine Users service is deprecated and '
    'will soon be removed. Please see '
    'https://cloud.google.com/appengine/docs/deprecations/open_id '
    'for details.')


class UserServiceStub(apiproxy_stub.APIProxyStub):
  """Trivial implementation of the `UserService`."""

  _ACCEPTS_REQUEST_ID = True

  THREADSAFE = True

  def __init__(self,
               login_url=_DEFAULT_LOGIN_URL,
               logout_url=_DEFAULT_LOGOUT_URL,
               service_name='user',
               auth_domain=_DEFAULT_AUTH_DOMAIN,
               request_data=None):
    """Initializer.

    Args:
      login_url: String containing the URL to use for logging in.
      logout_url: String containing the URL to use for logging out.
      service_name: Service name expected for all calls.
      auth_domain: The authentication domain for the service (e.g. "gmail.com").
      request_data: A `apiproxy_stub.RequestData` instance used to look up state
          associated with the request that generated an API call.

    Note: Both the `login_url` and `logout_url` arguments must contain one
    format parameter, which will be replaced with the continuation URL where
    the user should be redirected after log-in or log-out has been completed.
    """
    super(UserServiceStub, self).__init__(service_name,
                                          request_data=request_data)
    self._login_url = login_url
    self._logout_url = logout_url
    self.__scopes = None

    self.SetOAuthUser(is_admin=oauth_api._OAUTH_IS_ADMIN.get(False))




    ctx_test_util.set_both('AUTH_DOMAIN', auth_domain)

  def SetOAuthUser(self,
                   email=_OAUTH_EMAIL,
                   domain=_OAUTH_AUTH_DOMAIN,
                   user_id=_OAUTH_USER_ID,
                   is_admin=False,
                   scopes=None,
                   client_id=_OAUTH_CLIENT_ID):
    """Set test OAuth user.

    Determines what user is returned by requests to `GetOAuthUser`.

    Args:
      email: Email address of OAuth user.  None indicates that no oauth user
        is authenticated.
      domain: Domain of OAuth user.
      user_id: User ID of OAuth user.
      is_admin:  Whether the user is an admin.
      scopes: List of scopes that user is authenticated against.
      client_id: Client ID of the OAuth2 request
    """
    self.__email = email
    self.__domain = domain
    self.__user_id = user_id
    self.__is_admin = is_admin
    self.__scopes = scopes
    self.__client_id = client_id

  def _Dynamic_SetOAuthUser(self, request, unused_response, unused_request_id):
    """Local implementation of `UserStubService.SetOAuthUser()`.

    Args:
      request: A `user_stub_service_pb.SetOAuthUserRequest` message.
    """
    self.SetOAuthUser(
        email=request.email or self.__email,
        domain=request.auth_domain or self.__domain,
        user_id=request.user_id or self.__user_id,
        is_admin=request.is_admin or self.__is_admin,
        scopes=request.scopes or self.__scopes,
        client_id=request.client_id or self.__client_id)

  def _Dynamic_CreateLoginURL(self, request, response, request_id):
    """Trivial implementation of `UserService.CreateLoginURL()`.

    Args:
      request: A `CreateLoginURLRequest`.
      response: A `CreateLoginURLResponse`.
      request_id: A unique string identifying the request associated with the
          API call.
    """
    if request.HasField('federated_identity'):
      logging.warning(_OPENID_DEPRECATION_WARNING)

    response.login_url = self._login_url % urllib.parse.quote(
            self._AddHostToContinueURL(request.destination_url, request_id))

  def _Dynamic_CreateLogoutURL(self, request, response, request_id):
    """Trivial implementation of `UserService.CreateLogoutURL()`.

    Args:
      request: A `CreateLogoutURLRequest`.
      response: A `CreateLogoutURLResponse`.
      request_id: A unique string identifying the request associated with the
          API call.
    """
    response.logout_url = self._logout_url % urllib.parse.quote(self._AddHostToContinueURL(request.destination_url, request_id))

  def _Dynamic_GetOAuthUser(self, request, response, request_id):
    """Trivial implementation of `UserService.GetOAuthUser()`.

    Args:
      request: A `GetOAuthUserRequest`.
      response: A `GetOAuthUserResponse`.
      request_id: A unique string identifying the request associated with the
          API call.
    """
    if self.__email is None:
      raise apiproxy_errors.ApplicationError(
          user_service_pb2.UserServiceError.OAUTH_INVALID_REQUEST)
    else:
      if self.__scopes is None:
        authorized_scopes = set()
      else:

        authorized_scopes = set(request.scopes).intersection(
            self.__scopes)
        if not authorized_scopes:
          raise apiproxy_errors.ApplicationError(
              user_service_pb2.UserServiceError.OAUTH_INVALID_TOKEN)
      response.email = self.__email
      response.user_id = self.__user_id
      response.auth_domain = self.__domain
      response.is_admin = self.__is_admin
      response.client_id = self.__client_id
      response.scopes.extend(authorized_scopes)

  def _AddHostToContinueURL(self, continue_url, request_id):
    """Adds the request host to the continue URL if no host is specified.

    Args:
      continue_url: The URL which may or may not have a host specified
      request_id: A unique string identifying the request associated with the
          API call.

    Returns:
      String.
    """
    (protocol, host, path, parameters, query, fragment) = six.moves.urllib.parse.urlparse(continue_url)

    if host and protocol:
      return continue_url

    try:
      protocol, host, _, _, _, _ = six.moves.urllib.parse.urlparse(
          self.request_data.get_request_url(request_id))
    except KeyError:



      pass


    if path == '':
      path = '/'

    return six.moves.urllib.parse.urlunparse(
        (protocol, host, path, parameters, query, fragment))



#!/usr/bin/env python
#
# Copyright 2007 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

"""Performs translation of `queue.xml` to `queue.yaml`."""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

import sys

from xml.etree import ElementTree

from google.appengine.tools import xml_parser_utils
from google.appengine.tools.app_engine_config_exception import AppEngineConfigException

PUSH_QUEUE_TAGS = (
    'rate', 'bucket-size', 'max-concurrent-requests', 'target')

PUSH_QUEUE_RETRY_PARAMS = (
    'task-age-limit',
    'min-backoff-seconds',
    'max-backoff-seconds',
    'max-doublings')

RETRY_PARAMETER_TAGS = ('task-retry-limit',) + PUSH_QUEUE_RETRY_PARAMS

BAD_MODE_ERROR_MESSAGE = (
    'Mode, if specified, must be either push or pull'
    ' (defaults to push) for queue entries. Bad value'
    " '%s' in <queue> entry with name '%s'")

PULL_QUEUE_ERROR_MESSAGE = (
    'The element <%s> is not defined for pull '
    "queues; bad <queue> entry with name '%s'")
RETRY_PARAM_ERROR_MESSAGE = (
    'The element <%s> in <retry-parameters> is not '
    "defined for pull queue with name '%s'")


def GetQueueYaml(unused_application, queue_xml_str):
  queue_xml = QueueXmlParser().ProcessXml(queue_xml_str)
  return queue_xml.ToYaml()


class QueueXmlParser(object):
  """Provides logic for walking down XML tree and pulling data."""

  def ProcessXml(self, xml_str):
    """Parses XML string and returns object representation of relevant info.

    Args:
      xml_str: The XML string.
    Returns:
      A `QueueXml` object containing information about task queue
      specifications from the XML.
    Raises:
      AppEngineConfigException: In case of malformed XML or illegal inputs.
    """

    try:
      self.errors = []
      xml_root = ElementTree.fromstring(xml_str)

      if xml_parser_utils.GetTag(xml_root) != 'queue-entries':
        raise AppEngineConfigException('Root tag must be <queue-entries>')

      self.queue_xml = QueueXml()
      self.queue_xml.queues = []
      self.queue_xml.total_storage_limit = xml_parser_utils.GetChildNodeText(
          xml_root, 'total-storage-limit')
      for child in xml_parser_utils.GetNodes(xml_root, 'queue'):
        self.ProcessQueueNode(child)

      if self.errors:
        raise AppEngineConfigException('\n'.join(self.errors))

      return self.queue_xml

    except ElementTree.ParseError as e:
      raise AppEngineConfigException('Bad input -- not valid XML: %s' % e)

  def ProcessQueueNode(self, node):
    """Processes XML `<queue>` nodes into Queue objects.

    The following information is parsed out:

    - name
    - mode: can be either push or pull
    - retry-parameters: task-retry-limit

    For push queues only:

    - task-age-limit
    - min-backoff-seconds
    - max-back-off-seconds
    - max-doubling
    - bucket-size
    - max-concurrent-requests
    - rate: how often tasks are processed on this queue.
    - target: version of application on which tasks on this queue will be
        invoked.

    For pull queues only:

    - acl: access control list - lists user and writer email addresses.

    Args:
      node: Current `<queue>` XML node being processed.
    """

    name = xml_parser_utils.GetChildNodeText(node, 'name')
    if not name:
      self.errors.append('Must specify a name for each <queue> entry')
      return


    mode = xml_parser_utils.GetChildNodeText(node, 'mode', 'push')

    if mode not in ('push', 'pull'):
      self.errors.append(BAD_MODE_ERROR_MESSAGE % (mode, name))
      return
    if mode == 'pull':
      queue = PullQueue()
      queue.name = name
      self._ProcessPullQueueNode(node, queue)
    else:
      queue = PushQueue()
      queue.name = name
      self._ProcessPushQueueNode(node, queue)

    self.queue_xml.queues.append(queue)

  def _ProcessPushQueueNode(self, node, queue):
    if xml_parser_utils.GetChild(node, 'acl') is not None:
      self.errors.append(
          'The element <acl> is not defined for push '
          "queues; bad <queue> entry with name '%s'" % queue.name)
    for tag in PUSH_QUEUE_TAGS:
      field_name = tag.replace('-', '_')
      setattr(queue, field_name, xml_parser_utils.GetChildNodeText(node, tag))
    self._ProcessRetryParametersNode(node, queue)

  def _ProcessPullQueueNode(self, node, queue):
    """Populates PullQueue-specific fields from parsed XML."""
    for tag in PUSH_QUEUE_TAGS:
      if xml_parser_utils.GetChild(node, tag) is not None:
        self.errors.append(PULL_QUEUE_ERROR_MESSAGE % (tag, queue.name))

    acl_node = xml_parser_utils.GetChild(node, 'acl')

    if acl_node is not None:
      queue.acl = Acl()
      queue.acl.user_emails = [
          sub_node.text for sub_node in
          xml_parser_utils.GetNodes(acl_node, 'user-email')]
      queue.acl.writer_emails = [
          sub_node.text for sub_node in
          xml_parser_utils.GetNodes(acl_node, 'writer-email')]
    else:
      queue.acl = None

    self._ProcessRetryParametersNode(node, queue)

  def _ProcessRetryParametersNode(self, node, queue):
    """Pulls information out of <retry-parameters> node."""
    retry_parameters_node = xml_parser_utils.GetChild(
        node, 'retry-parameters')
    if retry_parameters_node is None:
      queue.retry_parameters = None
      return
    retry_parameters = RetryParameters()
    queue.retry_parameters = retry_parameters
    retry_parameters.task_retry_limit = xml_parser_utils.GetChildNodeText(
        retry_parameters_node, 'task-retry-limit')

    for tag in PUSH_QUEUE_RETRY_PARAMS:

      if xml_parser_utils.GetChild(retry_parameters_node, tag) is not None:
        if isinstance(queue, PullQueue):
          self.errors.append(RETRY_PARAM_ERROR_MESSAGE % (tag, queue.name))
        else:
          setattr(
              retry_parameters,
              tag.replace('-', '_'),
              xml_parser_utils.GetChildNodeText(retry_parameters_node, tag))


class QueueXml(object):

  def __init__(self):
    self.queues = []
    self.total_storage_limit = None

  def ToYaml(self):
    statements = []
    if self.total_storage_limit:
      statements.append('total_storage_limit: %s\n' % self.total_storage_limit)
    statements.append('queue:')
    for queue in self.queues:
      statements += queue.GetYamlStatementsList()

    return '\n'.join(statements) + '\n'


class Queue(object):

  def GetYamlStatementsList(self):
    statements = ['- name: %s' % self.name]
    statements += self.GetAdditionalYamlStatementsList()

    if self.retry_parameters:
      statements += self.retry_parameters.GetYamlStatementsList()

    return statements


class PushQueue(Queue):

  def GetAdditionalYamlStatementsList(self):
    statements = ['  mode: push']
    fields = (tag.replace('-', '_') for tag in PUSH_QUEUE_TAGS)
    for field in fields:
      field_value = getattr(self, field)
      if field_value:
        statements.append('  %s: %s' % (field, field_value))

    return statements


class PullQueue(Queue):

  def GetAdditionalYamlStatementsList(self):
    statements = ['  mode: pull']
    if self.acl:
      statements += self.acl.GetYamlStatementsList()
    return statements


class Acl(object):

  def GetYamlStatementsList(self):
    statements = ['  acl:']
    statements += ['  - user_email: %s' % user_email for user_email in
                   self.user_emails]
    statements += ['  - writer_email: %s' % writer_email for writer_email in
                   self.writer_emails]
    return statements


class RetryParameters(object):

  def GetYamlStatementsList(self):
    statements = ['  retry_parameters:']
    field_names = (tag.replace('-', '_') for tag in RETRY_PARAMETER_TAGS)
    for field in field_names:
      field_value = getattr(self, field, None)
      if field_value:
        statements.append('    %s: %s' % (field, field_value))
    return statements


def main():
  xml = sys.stdin.read()
  print(GetQueueYaml(None, xml))

if __name__ == '__main__':
  sys.exit(main())



#!/usr/bin/env python
#
# Copyright 2007 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
"""Helper for users of the public App Engine SDK.

This is for use as part of the code at:
https://github.com/GoogleCloudPlatform/appengine-python-standard

Example for a standard WSGI app:

  ```python
  import google.appengine.api

  app = google.appengine.api.wrap_wsgi_app(app)
  ```

Example for a Flask app:

  ```python
  import google.appengine.api
  from flask import Flask, request

  app = Flask(__name__)
  app.wsgi_app = google.appengine.api.wrap_wsgi_app(app.wsgi_app)
  ```
"""
import os
from google.appengine.api import full_app_id




def wrap_wsgi_app(app, use_legacy_context_mode=True, use_deferred=False):
  """Wrap a WSGI app with middlewares required to access App Engine APIs."""


  from google.appengine.runtime import initialize
  from google.appengine.runtime import middlewares
  from google.appengine.runtime import default_api_stub




  initialize.InitializeThreadingApis()

  default_api_stub.Register(default_api_stub.DefaultApiStub())



  full_app_id.normalize()

  def if_legacy(array):
    return array if use_legacy_context_mode else []

  def if_deferred_enabled(array):
    return array if use_deferred else []

  return middlewares.Wrap(
      app,
      if_legacy([
          middlewares.MakeInitLegacyRequestOsEnvironMiddleware(),
      ]) + [
          middlewares.RunInNewContextMiddleware,
          middlewares.SetContextFromHeadersMiddleware,
          middlewares.CallbackMiddleware,
          middlewares.UseRequestSecurityTicketForApiMiddleware,
          middlewares.WaitForResponseMiddleware,
          middlewares.WsgiEnvSettingMiddleware,

          middlewares.MakeLegacyWsgiEnvSettingMiddleware(),
      ] + if_legacy([
          middlewares.LegacyWsgiRemoveXAppenginePrefixMiddleware,
          middlewares.LegacyCopyWsgiEnvToOsEnvMiddleware,
      ]) + [
          middlewares.ErrorLoggingMiddleware,
          middlewares.BackgroundAndShutdownMiddleware,
          middlewares.SetNamespaceFromHeader,
      ] + if_deferred_enabled([
          middlewares.AddDeferredMiddleware,
      ]))



#!/usr/bin/env python
#
# Copyright 2007 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
"""Save X-Appengine-* headers into request context."""
import contextvars
from typing import Dict

PREFIX = 'HTTP_X_APPENGINE_'

AUTH_DOMAIN = contextvars.ContextVar('AUTH_DOMAIN')
DEFAULT_VERSION_HOSTNAME = contextvars.ContextVar('DEFAULT_VERSION_HOSTNAME')
USER_EMAIL = contextvars.ContextVar('USER_EMAIL')
USER_ID = contextvars.ContextVar('USER_ID')
USER_IS_ADMIN = contextvars.ContextVar('USER_IS_ADMIN')
USER_NICKNAME = contextvars.ContextVar('USER_NICKNAME')
DEFAULT_NAMESPACE = contextvars.ContextVar('DEFAULT_NAMESPACE')


def init_from_wsgi_environ(
    wsgi_env) -> Dict[contextvars.ContextVar, contextvars.Token]:
  """Init contextvars from matching X_APPENGINE_ headers if found."""

  reset_tokens: Dict[contextvars.ContextVar, contextvars.Token] = {}
  for ctxvar in [v for _, v in globals().items()
                 if isinstance(v, contextvars.ContextVar)]:
    value = wsgi_env.get(PREFIX + ctxvar.name)
    if value is not None:
      if ctxvar.name == 'USER_IS_ADMIN':

        value = (value == '1')
      token = ctxvar.set(value)
      reset_tokens[ctxvar] = token
  return reset_tokens



#!/usr/bin/env python
#
# Copyright 2007 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#


"""Synchronization primitives backed by the datastore.

The `Mutex` class's interface matches the Python standard library's
`threading.Lock` class.

"""










import datetime
import random
import time

from google.appengine.ext import db
from google.appengine.api import datastore


class Mutex(db.Model):
  """A mutex backed by the datastore."""
  LIFETIME = datetime.timedelta(milliseconds=20*1000)

  _held = db.BooleanProperty(required=True, default=False)
  _last_acquired = db.DateTimeProperty(required=True, auto_now=True)

  def acquire(self, blocking=True, retry_delay_ms=100):
    """Acquire this mutex. Fails if it's already held.

    Args:
      blocking: boolean. If `True`, blocks until the mutex can be acquired. If
      `False`, returns immediately.
      retry_delay_ms: integer. The desired delay between acquisition retries.
      Each actually delay is a random amount between half and twice this.

    Returns:
      `True` if the mutex was acquired, `False` otherwise.
    """
    if not self.is_saved():
      self.put()

    while True:
      mutex = db.run_in_transaction(self._acquire_or_fail)
      if mutex is not None:

        self._held = True
        self._last_acquired = mutex._last_acquired
        return True
      elif not blocking:
        return False
      else:
        retry_delay_s = float(retry_delay_ms) / 1000
        time.sleep(random.uniform(retry_delay_s / 2, retry_delay_s * 2))

  def release(self):
    """Releases the mutex. Raises `db.BadRequestError` if not currently held."""
    if not self.is_saved():
      raise db.BadRequestError("Can't release mutex that's not saved.")

    db.run_in_transaction(self._release_or_fail)


    self._held = False

  def is_held(self):
    """Returns `True` if the mutex is held, `False` otherwise.

    This takes expiration into account.
    """
    elapsed = datetime.datetime.now() - self._last_acquired
    return self._held and elapsed < Mutex.LIFETIME

  def _acquire_or_fail(self):
    """Acquires the given mutex.

    Intended to be run in a transaction.

    Returns:
      The mutex if it was acquired, otherwise `None`.
    """
    mutex = Mutex.get(self.key())
    if not mutex.is_held():
      mutex._held = True
      mutex.put()
      return mutex
    else:
      return None

  def _release_or_fail(self):
    """Releases the given mutex.

    Raises `db.BadRequestError` if it's held.

    Intended to be run in a transaction.

    Returns:
      Mutex
    """
    mutex = Mutex.get(self.key())
    if mutex.is_held():
      mutex._held = False
      mutex.put()
      return mutex
    else:
      raise db.BadRequestError("Can't release mutex that's not held.")


def acquire(mutexes, blocking=True):
  """Acquires the given mutexes.

  To prevent deadlocks, mutexes are acquired in order of the partial ordering
  defined by their keys.

  Args:
    mutexes: Sequence of Mutex objects or keys
    blocking: Boolean. If `True`, blocks until all mutexes can be acquired. If
      `False`, returns immediately.

  Returns:
    `True` if all mutexes were acquired, `False` otherwise.
  """

  mutexes = _to_mutexes(mutexes)
  mutexes.sort(key=lambda m: m.key())

  acquired = []

  try:
    for mutex in mutexes:
      if mutex.acquire(blocking=blocking):
        acquired.append(mutex)
      else:
        return False
  finally:
    if len(acquired) < len(mutexes):
      release(acquired)

  return True

def release(mutexes):
  """Releases the given mutexes.

  Mutexes are released according to the reverse of the partial ordering
  defined by their keys. This isn't strictly necessary to prevent deadlocks,
  but it's still a nice practice.

  Args:
    mutexes: sequence of Mutex objects or keys
  """
  mutexes = _to_mutexes(mutexes)
  mutexes.sort(key=lambda m: m.key(), reverse=True)

  for mutex in mutexes:
    mutex.release()


def _to_mutexes(mutexes_or_keys):
  """Normalizes and type checks the given sequence.

  Args:
    mutexes_or_keys: sequence of Mutex objects or keys

  Returns:
    list of Mutexes
  """
  mutexes = []

  for arg in mutexes_or_keys:
    if isinstance(arg, Mutex):
      mutexes.append(arg)
    elif isinstance(arg, db.Key):
      mutexes.append(Mutex.get(arg))
    else:
      raise db.BadArgumentError(
          'Expected a Mutex instance or key; received %s (a %s).' %
          (arg, datastore.typename(arg)))

  return mutexes



#!/usr/bin/env python
#
# Copyright 2007 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#



"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database


_sym_db = _symbol_database.Default()




DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n*google/appengine/datastore/entity_v4.proto\x12\x1dgoogle.appengine.datastore.v4\"X\n\x0bPartitionId\x12\x12\n\ndataset_id\x18\x03 \x01(\t\x12\x11\n\tnamespace\x18\x04 \x01(\t\"\"\n\tConstants\x12\x15\n\x11MAX_DIMENSION_TAG\x10\x64\"\xc4\x01\n\x03Key\x12@\n\x0cpartition_id\x18\x01 \x01(\x0b\x32*.google.appengine.datastore.v4.PartitionId\x12\x44\n\x0cpath_element\x18\x02 \x03(\x0b\x32..google.appengine.datastore.v4.Key.PathElement\x1a\x35\n\x0bPathElement\x12\x0c\n\x04kind\x18\x01 \x02(\t\x12\n\n\x02id\x18\x02 \x01(\x03\x12\x0c\n\x04name\x18\x03 \x01(\t\"/\n\x08GeoPoint\x12\x10\n\x08latitude\x18\x01 \x02(\x01\x12\x11\n\tlongitude\x18\x02 \x02(\x01\"\xcb\x03\n\x05Value\x12\x15\n\rboolean_value\x18\x01 \x01(\x08\x12\x15\n\rinteger_value\x18\x02 \x01(\x03\x12\x14\n\x0c\x64ouble_value\x18\x03 \x01(\x01\x12$\n\x1ctimestamp_microseconds_value\x18\x04 \x01(\x03\x12\x35\n\tkey_value\x18\x05 \x01(\x0b\x32\".google.appengine.datastore.v4.Key\x12\x16\n\x0e\x62lob_key_value\x18\x10 \x01(\t\x12\x14\n\x0cstring_value\x18\x11 \x01(\t\x12\x12\n\nblob_value\x18\x12 \x01(\x0c\x12;\n\x0c\x65ntity_value\x18\x06 \x01(\x0b\x32%.google.appengine.datastore.v4.Entity\x12@\n\x0fgeo_point_value\x18\x08 \x01(\x0b\x32\'.google.appengine.datastore.v4.GeoPoint\x12\x38\n\nlist_value\x18\x07 \x03(\x0b\x32$.google.appengine.datastore.v4.Value\x12\x0f\n\x07meaning\x18\x0e \x01(\x05\x12\x15\n\x07indexed\x18\x0f \x01(\x08:\x04true\"\xb6\x01\n\x08Property\x12\x0c\n\x04name\x18\x01 \x02(\t\x12#\n\x10\x64\x65precated_multi\x18\x02 \x01(\x08:\x05\x66\x61lseB\x02\x18\x01\x12\x42\n\x10\x64\x65precated_value\x18\x03 \x03(\x0b\x32$.google.appengine.datastore.v4.ValueB\x02\x18\x01\x12\x33\n\x05value\x18\x04 \x01(\x0b\x32$.google.appengine.datastore.v4.Value\"t\n\x06\x45ntity\x12/\n\x03key\x18\x01 \x01(\x0b\x32\".google.appengine.datastore.v4.Key\x12\x39\n\x08property\x18\x02 \x03(\x0b\x32\'.google.appengine.datastore.v4.PropertyB\'\n%com.google.google.appengine.datastore')



_PARTITIONID = DESCRIPTOR.message_types_by_name['PartitionId']
_KEY = DESCRIPTOR.message_types_by_name['Key']
_KEY_PATHELEMENT = _KEY.nested_types_by_name['PathElement']
_GEOPOINT = DESCRIPTOR.message_types_by_name['GeoPoint']
_VALUE = DESCRIPTOR.message_types_by_name['Value']
_PROPERTY = DESCRIPTOR.message_types_by_name['Property']
_ENTITY = DESCRIPTOR.message_types_by_name['Entity']
_PARTITIONID_CONSTANTS = _PARTITIONID.enum_types_by_name['Constants']
PartitionId = _reflection.GeneratedProtocolMessageType('PartitionId', (_message.Message,), {
  'DESCRIPTOR' : _PARTITIONID,
  '__module__' : 'google.appengine.datastore.entity_v4_pb2'

  })
_sym_db.RegisterMessage(PartitionId)

Key = _reflection.GeneratedProtocolMessageType('Key', (_message.Message,), {

  'PathElement' : _reflection.GeneratedProtocolMessageType('PathElement', (_message.Message,), {
    'DESCRIPTOR' : _KEY_PATHELEMENT,
    '__module__' : 'google.appengine.datastore.entity_v4_pb2'

    })
  ,
  'DESCRIPTOR' : _KEY,
  '__module__' : 'google.appengine.datastore.entity_v4_pb2'

  })
_sym_db.RegisterMessage(Key)
_sym_db.RegisterMessage(Key.PathElement)

GeoPoint = _reflection.GeneratedProtocolMessageType('GeoPoint', (_message.Message,), {
  'DESCRIPTOR' : _GEOPOINT,
  '__module__' : 'google.appengine.datastore.entity_v4_pb2'

  })
_sym_db.RegisterMessage(GeoPoint)

Value = _reflection.GeneratedProtocolMessageType('Value', (_message.Message,), {
  'DESCRIPTOR' : _VALUE,
  '__module__' : 'google.appengine.datastore.entity_v4_pb2'

  })
_sym_db.RegisterMessage(Value)

Property = _reflection.GeneratedProtocolMessageType('Property', (_message.Message,), {
  'DESCRIPTOR' : _PROPERTY,
  '__module__' : 'google.appengine.datastore.entity_v4_pb2'

  })
_sym_db.RegisterMessage(Property)

Entity = _reflection.GeneratedProtocolMessageType('Entity', (_message.Message,), {
  'DESCRIPTOR' : _ENTITY,
  '__module__' : 'google.appengine.datastore.entity_v4_pb2'

  })
_sym_db.RegisterMessage(Entity)

if _descriptor._USE_C_DESCRIPTORS == False:

  DESCRIPTOR._options = None
  DESCRIPTOR._serialized_options = b'\n%com.google.google.appengine.datastore'
  _PROPERTY.fields_by_name['deprecated_multi']._options = None
  _PROPERTY.fields_by_name['deprecated_multi']._serialized_options = b'\030\001'
  _PROPERTY.fields_by_name['deprecated_value']._options = None
  _PROPERTY.fields_by_name['deprecated_value']._serialized_options = b'\030\001'
  _PARTITIONID._serialized_start=77
  _PARTITIONID._serialized_end=165
  _PARTITIONID_CONSTANTS._serialized_start=131
  _PARTITIONID_CONSTANTS._serialized_end=165
  _KEY._serialized_start=168
  _KEY._serialized_end=364
  _KEY_PATHELEMENT._serialized_start=311
  _KEY_PATHELEMENT._serialized_end=364
  _GEOPOINT._serialized_start=366
  _GEOPOINT._serialized_end=413
  _VALUE._serialized_start=416
  _VALUE._serialized_end=875
  _PROPERTY._serialized_start=878
  _PROPERTY._serialized_end=1060
  _ENTITY._serialized_start=1062
  _ENTITY._serialized_end=1178




#!/usr/bin/env python
#
# Copyright 2007 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
"""Minimal library for communicating with the metadata server.

Branched from
http://google/third_party/py/google/auth/compute_engine/_metadata.py
and enhanced to support explicit oauth2 scopes.
"""



import os
import time
from typing import List, Optional, Text, Tuple

import frozendict
import requests

_METADATA_ROOT = 'http://{}/computeMetadata/v1/'.format(
    os.getenv('GCE_METADATA_HOST', 'metadata.google.internal'))

_METADATA_HEADERS = frozendict.frozendict({'metadata-flavor': 'Google'})


class TransportError(Exception):
  """Metadata server response not 200 OK."""


def _get(path: Text, root: Text = _METADATA_ROOT, **kwargs):
  kwargs.setdefault('headers', _METADATA_HEADERS)
  url = root + path
  response = requests.get(url, **kwargs)
  if response.status_code != requests.codes.ok:
    raise TransportError(
        'Failed to retrieve {} from the metadata service. '
        'Status: {} Response:\n{}'.format(
            url, response.status_code, response.content))
  return response.json()


def get_service_account_token(
    scopes: List[Text],
    service_account: Optional[Text] = None) -> Tuple[Text, int]:
  """Get the OAuth 2.0 access token for a service account.

  Args:
      scopes: A list of scopes to be sent with the credentials.
      service_account: The string 'default' or a service account email
          address. The determines which service account for which to acquire
          an access token.

  Returns:
      The access token and its expiration in epoch time.

  Raises:
      TransportError: if an error occurred while
          retrieving metadata.
  """
  if service_account is None:
    service_account = 'default'

  token_json = _get(
      path='instance/service-accounts/{0}/token'.format(service_account),
      params={'scopes': ','.join(scopes)}
  )
  return token_json['access_token'], time.time()+token_json['expires_in']



#!/usr/bin/env python
#
# Copyright 2007 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#


"""Control the namespacing system used by various APIs.

A namespace may be specified in various API calls exemplified
by the datastore and memcache interfaces. The default can be
specified using this module.
"""







import contextvars
import os
import re

from google.appengine.runtime import context

__all__ = ['BadValueError',
           'set_namespace',
           'get_namespace',
           'google_apps_namespace',
           'validate_namespace',
          ]




_CURRENT_NAMESPACE = contextvars.ContextVar('_CURRENT_NAMESPACE')
_TESTBED_RESET_TOKEN = None






_NAMESPACE_MAX_LENGTH = 100
_NAMESPACE_PATTERN = r'^[0-9A-Za-z._-]{0,%s}$' % _NAMESPACE_MAX_LENGTH
_NAMESPACE_RE = re.compile(_NAMESPACE_PATTERN)


def set_namespace(namespace):
  """Set the default namespace for the current HTTP request.

  Args:
    namespace: A string naming the new namespace to use. A value of None
      will unset the default namespace value.
  """
  if namespace is not None:
    validate_namespace(namespace)
  else:
    namespace = ''
  token = _CURRENT_NAMESPACE.set(namespace)
  global _TESTBED_RESET_TOKEN
  if _TESTBED_RESET_TOKEN is None:
    _TESTBED_RESET_TOKEN = token


def get_namespace():
  """Get the current default namespace or (`''`) namespace if unset."""
  return _CURRENT_NAMESPACE.get('')


def google_apps_namespace():
  if context.READ_FROM_OS_ENVIRON:
    return os.environ.get('HTTP_X_APPENGINE_DEFAULT_NAMESPACE')
  else:
    return context.gae_headers.DEFAULT_NAMESPACE.get()


class BadValueError(Exception):
  """Raised by `ValidateNamespaceString`."""


def validate_namespace(value, exception=BadValueError):
  """Raises an exception if value is not a valid namespace string.

  A namespace must match the regular expression `([0-9A-Za-z._-]{0,100})`.

  Args:
    value: string, the value to validate.
    exception: exception type to raise.

  Raises:
    BadValueError: If value is not a valid namespace string.
  """
  if not isinstance(value, str):
    raise exception(
        'value should be a string; received %r (a %s):' % (value, type(value)))
  if not _NAMESPACE_RE.match(value):
    raise exception(
        'value "%s" does not match regex "%s"' % (value, _NAMESPACE_PATTERN))
